class StartingGridEntry:
    """
    Represents an entry on the starting grid.
    """
    def __init__(self, position, driver_index, driver_name):
        self._position = position
        self._driver_index = driver_index
        self._driver_name = driver_name

    @property
    def driver_index(self):
        """
        Index position of driver.
        """
        return self._driver_index

    @property
    def driver_name(self):
        """
        Telemetry-read name of driver.
        """
        return self._driver_name

    @property
    def position(self):
        """
        Starting position of driver.
        """
        return self._position



"""
Tests RaceData.py.
"""
import sys
import unittest
from unittest.mock import MagicMock, PropertyMock, patch, sentinel

from replayenhancer.RaceData import RaceData, Driver, \
    ClassificationEntry, SectorTime


class TestRaceData(unittest.TestCase):
    """
    Tests against the RaceData object.
    """
    instance = None

    def setUp(self):
        with patch('replayenhancer.RaceData.TelemetryData') \
                as mock_telemetry_data, \
                patch('replayenhancer.RaceData.RaceData.get_data') \
                as mock_get_data:
            mock_telemetry_data.return_value = sentinel.telemetry_data
            mock_get_data.return_value = sentinel.get_data
            self.instance = RaceData(sentinel.directory)

        mock_driver_1 = MagicMock(spec=Driver)
        type(mock_driver_1).best_lap = PropertyMock(return_value=24.88)
        type(mock_driver_1).best_sector_1 = PropertyMock(return_value=2.83)
        type(mock_driver_1).best_sector_2 = PropertyMock(return_value=13.64)
        type(mock_driver_1).best_sector_3 = PropertyMock(return_value=8.37)
        type(mock_driver_1).index = PropertyMock(return_value=1)

        mock_driver_2 = MagicMock(spec=Driver)
        type(mock_driver_2).best_lap = PropertyMock(return_value=29.77)
        type(mock_driver_2).best_sector_1 = PropertyMock(return_value=3.12)
        type(mock_driver_2).best_sector_2 = PropertyMock(return_value=15.93)
        type(mock_driver_2).best_sector_3 = PropertyMock(return_value=10.67)
        type(mock_driver_2).index = PropertyMock(return_value=2)

        mock_driver_3 = MagicMock(spec=Driver)
        type(mock_driver_3).best_lap = PropertyMock(return_value=30.12)
        type(mock_driver_3).best_sector_1 = PropertyMock(return_value=3.15)
        type(mock_driver_3).best_sector_2 = PropertyMock(return_value=16.23)
        type(mock_driver_3).best_sector_3 = PropertyMock(return_value=10.70)
        type(mock_driver_3).index = PropertyMock(return_value=None)

        self.driver_data = {
            "Kobernulf Monnur": mock_driver_1,
            "Scott Winstead": mock_driver_2,
            "Timon Putzkerd": mock_driver_3,
            "Timon Putzkernur": mock_driver_3
        }
        self.instance.drivers = self.driver_data

    def tearDown(self):
        self.instance = None

    def test_init(self):
        expected_result = RaceData
        self.assertIsInstance(self.instance, expected_result)

    def test_init_directory(self):
        expected_result = sentinel.directory
        self.assertEqual(self.instance._telemetry_directory, expected_result)

    def test_init_descriptor(self):
        expected_result = "descriptor.json"
        self.assertEqual(self.instance._descriptor_filename, expected_result)

    def test_init_custom_descriptor(self):
        with patch('replayenhancer.RaceData.TelemetryData') \
                as mock_telemetry_data, \
                patch('replayenhancer.RaceData.RaceData.get_data') \
                as mock_get_data:
            mock_telemetry_data.return_value = sentinel.telemetry_data
            mock_get_data.return_value = sentinel.get_data
            instance = RaceData(sentinel.directory, descriptor_filename=sentinel.descriptor)

        expected_result = sentinel.descriptor
        self.assertEqual(instance._descriptor_filename, expected_result)

    def test_property_best_lap(self):
        expected_result = 24.88
        self.assertEqual(self.instance.best_lap, expected_result)

    def test_property_best_lap_no_laps(self):
        self.instance.drivers = dict()
        self.assertIsNone(self.instance.best_lap)

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_property_best_lap_error(self):
        with patch('replayenhancer.RaceData.min') as mock_min:
            mock_min.side_effect = ValueError
            self.assertIsNone(self.instance.best_lap)

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_property_best_lap_error_pre35(self):
        with patch('builtins.min') as mock_min:
            mock_min.side_effect = ValueError
            self.assertIsNone(self.instance.best_lap)

    def test_property_best_sector_1(self):
        expected_result = 2.83
        self.assertEqual(self.instance.best_sector_1, expected_result)

    def test_property_best_sector_1_no_laps(self):
        self.instance.drivers = dict()
        self.assertIsNone(self.instance.best_sector_1)

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_property_best_sector_1_error(self):
        with patch('replayenhancer.RaceData.min') as mock_min:
            mock_min.side_effect = ValueError
            self.assertIsNone(self.instance.best_sector_1)

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_property_best_sector_1_error_pre35(self):
        with patch('builtins.min') as mock_min:
            mock_min.side_effect = ValueError
            self.assertIsNone(self.instance.best_sector_1)

    def test_property_best_sector_2(self):
        expected_result = 13.64
        self.assertEqual(self.instance.best_sector_2, expected_result)

    def test_property_best_sector_2_no_laps(self):
        self.instance.drivers = dict()
        self.assertIsNone(self.instance.best_sector_2)

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_property_best_sector_2_error(self):
        with patch('replayenhancer.RaceData.min') as mock_min:
            mock_min.side_effect = ValueError
            self.assertIsNone(self.instance.best_sector_2)

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_property_best_sector_2_error_pre35(self):
        with patch('builtins.min') as mock_min:
            mock_min.side_effect = ValueError
            self.assertIsNone(self.instance.best_sector_2)

    def test_property_best_sector_3(self):
        expected_result = 8.37
        self.assertEqual(self.instance.best_sector_3, expected_result)

    def test_property_best_sector_3_no_laps(self):
        self.instance.drivers = dict()
        self.assertIsNone(self.instance.best_sector_3)

    @unittest.skipIf(sys.version_info < (3, 5), "Not supported.")
    def test_property_best_sector_3_error(self):
        with patch('replayenhancer.RaceData.min') as mock_min:
            mock_min.side_effect = ValueError
            self.assertIsNone(self.instance.best_sector_3)

    @unittest.skipIf(sys.version_info >= (3, 5), "Not supported.")
    def test_property_best_sector_3_error_pre35(self):
        with patch('builtins.min') as mock_min:
            mock_min.side_effect = ValueError
            self.assertIsNone(self.instance.best_sector_3)

    @unittest.skip("I'm not smart enough to do this.")
    @patch('replayenhancer.RaceData.TelemetryData', autospec=True)
    def test_property_classification(self, mock_telemetry):
        mock_telemetry.return_value = sentinel.telemetry_data

        instance = RaceData(sentinel.directory)
        expected_result = list
        self.assertIsInstance(instance.classification, expected_result)

    def test_property_current_lap(self):
        mock_driver_1 = MagicMock(spec='replayenhancer.RETelemetryDataPacket.REParticipantInfo')
        type(mock_driver_1).current_lap = PropertyMock(return_value=3)

        mock_driver_2 = MagicMock(spec='replayenhancer.RETelemetryDataPacket.REParticipantInfo')
        type(mock_driver_2).current_lap = PropertyMock(return_value=3)

        mock_driver_3 = MagicMock(spec='replayenhancer.RETelemetryDataPacket.REParticipantInfo')
        type(mock_driver_3).current_lap = PropertyMock(return_value=2)

        mock_driver_4 = MagicMock(spec='replayenhancer.RETelemetryDataPacket.REParticipantInfo')
        type(mock_driver_4).current_lap = PropertyMock(return_value=1)

        self.instance._next_packet = MagicMock('replayenhancer.RETelemetryDataPacket.RETelemetryDataPacket', laps_in_event=10, participant_info=[
            mock_driver_1, mock_driver_2, mock_driver_3, mock_driver_4])

        expected_value = 3
        self.assertEqual(self.instance.current_lap, expected_value)

    def test_property_current_lap_overflow(self):
        mock_driver_1 = MagicMock(spec='replayenhancer.RETelemetryDataPacket.REParticipantInfo')
        type(mock_driver_1).current_lap = PropertyMock(return_value=11)

        mock_driver_2 = MagicMock(spec='replayenhancer.RETelemetryDataPacket.REParticipantInfo')
        type(mock_driver_2).current_lap = PropertyMock(return_value=10)

        mock_driver_3 = MagicMock(spec='replayenhancer.RETelemetryDataPacket.REParticipantInfo')
        type(mock_driver_3).current_lap = PropertyMock(return_value=6)

        mock_driver_4 = MagicMock(spec='replayenhancer.RETelemetryDataPacket.REParticipantInfo')
        type(mock_driver_4).current_lap = PropertyMock(return_value=5)

        self.instance._next_packet = MagicMock('replayenhancer.RETelemetryDataPacket.RETelemetryDataPacket', laps_in_event=10, participant_info=[
            mock_driver_1, mock_driver_2, mock_driver_3, mock_driver_4])

        expected_value = 10
        self.assertEqual(self.instance.current_lap, expected_value)

    def test_property_elapsed_time(self):
        expected_result = 0.0
        self.assertEqual(self.instance.elapsed_time, expected_result)

    def test_property_telemetry_data(self):
        expected_result = sentinel.telemetry_data
        self.assertEqual(self.instance.telemetry_data, expected_result)

    def test_property_total_laps(self):
        self.instance._next_packet = MagicMock('replayenhancer.RETelemetryDataPacket.RETelemetryDataPacket', laps_in_event=10)
        expected_value = 10
        self.assertEqual(self.instance.total_laps, expected_value)

    @unittest.skip("I'm not smart enough to do this.")
    def test_method_get_data(self):
        pass


class TestClassificationEntry(unittest.TestCase):
    """
    Unit tests for ClassificationEntry object.
    """
    instance = None

    def setUp(self):
        mock_driver = MagicMock(
            spec=Driver,
            best_lap=24.88,
            best_sector_1=2.83,
            best_sector_2=13.64,
            best_sector_3=8.37,
            laps_complete=3,
            race_time=181.00)
        mock_driver.configure_mock(name="Kobernulf Monnur")
        self.instance = ClassificationEntry(1, mock_driver, True)

    def tearDown(self):
        self.instance = None

    def test_init(self):
        expected_result = ClassificationEntry
        self.assertIsInstance(self.instance, expected_result)

    def test_property_best_lap(self):
        expected_result = 24.88
        self.assertEqual(self.instance.best_lap, expected_result)

    def test_property_best_sector_1(self):
        expected_result = 2.83
        self.assertEqual(self.instance.best_sector_1, expected_result)

    def test_property_best_sector_2(self):
        expected_result = 13.64
        self.assertEqual(self.instance.best_sector_2, expected_result)

    def test_property_best_sector_3(self):
        expected_result = 8.37
        self.assertEqual(self.instance.best_sector_3, expected_result)

    def test_property_driver(self):
        expected_result = Driver
        self.assertIsInstance(self.instance.driver, expected_result)

    def test_property_driver_name(self):
        expected_result = "Kobernulf Monnur"
        self.assertEqual(self.instance.driver_name, expected_result)

    def test_property_viewed_driver(self):
        expected_result = True
        self.assertEqual(self.instance.viewed_driver, expected_result)

    def test_property_laps_complete(self):
        expected_result = 3
        self.assertEqual(self.instance.laps_complete, expected_result)

    def test_property_calc_points_data(self):
        expected_result = ("Kobernulf Monnur", 1, 24.88)
        self.assertTupleEqual(self.instance.calc_points_data, expected_result)

    def test_property_position(self):
        expected_result = 1
        self.assertEqual(self.instance.position, expected_result)

    def test_property_race_time(self):
        expected_result = 181.00
        self.assertEqual(self.instance.race_time, expected_result)


class TestDriver(unittest.TestCase):
    """
    Unit tests for Driver object.
    """

    def test_init(self):
        instance = Driver(sentinel.index, sentinel.name)
        expected_result = Driver
        self.assertIsInstance(instance, expected_result)

    def test_property_best_lap_default(self):
        instance = Driver(sentinel.index, sentinel.name)
        self.assertIsNone(instance.best_lap)

    def test_property_best_lap_valid(self):
        instance = Driver(sentinel.index, sentinel.name)
        test_data = [
            (30.000, 1, False),
            (45.000, 2, False),
            (40.000, 3, False),
            (35.000, 1, False),
            (43.000, 2, False),
            (39.000, 3, False)
        ]
        expected_value = 115.000
        for data in test_data:
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(*data)
                sector_time.time = data[0]
                sector_time.sector = data[1]
                sector_time.invalid = data[2]
                instance.add_sector_time(sector_time)

        self.assertEqual(instance.best_lap, expected_value)

    def test_property_best_lap_invalid(self):
        instance = Driver(sentinel.index, sentinel.name)
        test_data = [
            (30.000, 1, False),
            (45.000, 2, False),
            (40.000, 3, True),
            (35.000, 1, False),
            (43.000, 2, False),
            (39.000, 3, False)
        ]
        expected_value = 117.000
        for data in test_data:
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(*data)
                sector_time.time = data[0]
                sector_time.sector = data[1]
                sector_time.invalid = data[2]
                instance.add_sector_time(sector_time)

        self.assertEqual(instance.best_lap, expected_value)

    def test_property_best_lap_all_invalid(self):
        instance = Driver(sentinel.index, sentinel.name)
        test_data = [
            (30.000, 1, False),
            (45.000, 2, False),
            (40.000, 3, True),
            (35.000, 1, True),
            (43.000, 2, False),
            (39.000, 3, False)
        ]
        for data in test_data:
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(*data)
                sector_time.time = data[0]
                sector_time.sector = data[1]
                sector_time.invalid = data[2]
                instance.add_sector_time(sector_time)

        self.assertIsNone(instance.best_lap)

    def test_property_best_lap_uneven(self):
        instance = Driver(sentinel.index, sentinel.name)
        test_data = [
            (30.000, 1, False),
            (45.000, 2, False),
            (40.000, 3, False),
            (35.000, 1, False),
            (43.000, 2, False)
        ]
        expected_value = 115.000

        for data in test_data:
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(*data)
                sector_time.time = data[0]
                sector_time.sector = data[1]
                sector_time.invalid = data[2]
                instance.add_sector_time(sector_time)

        self.assertEqual(instance.best_lap, expected_value)

    def test_property_best_lap_bad_first_sector(self):
        instance = Driver(sentinel.index, sentinel.name)
        test_data = [
            (16.000, 3, False),
            (30.000, 1, False),
            (45.000, 2, False),
            (40.000, 3, False),
            (35.000, 1, False),
            (43.000, 2, False),
            (39.000, 3, False)
        ]
        expected_value = 115.000

        for data in test_data:
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(*data)
                sector_time.time = data[0]
                sector_time.sector = data[1]
                sector_time.invalid = data[2]
                instance.add_sector_time(sector_time)

        self.assertEqual(instance.best_lap, expected_value)

    def test_property_best_sector_1_default(self):
        instance = Driver(sentinel.index, sentinel.name)
        self.assertIsNone(instance.best_sector_1)

    def test_property_best_sector_1_valid(self):
        instance = Driver(sentinel.index, sentinel.name)
        test_data = [
            (30.000, 1, False),
            (45.000, 2, False),
            (40.000, 3, False),
            (33.000, 1, False),
            (43.000, 2, False),
            (39.000, 3, False)
        ]
        expected_value = 30.000
        for data in test_data:
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(*data)
                sector_time.time = data[0]
                sector_time.sector = data[1]
                sector_time.invalid = data[2]
                instance.add_sector_time(sector_time)

        self.assertEqual(instance.best_sector_1, expected_value)

    def test_property_best_sector_1_invalid(self):
        instance = Driver(sentinel.index, sentinel.name)
        test_data = [
            (30.000, 1, True),
            (45.000, 2, False),
            (40.000, 3, False),
            (33.000, 1, False),
            (43.000, 2, False),
            (39.000, 3, False)
        ]
        expected_value = 33.000
        for data in test_data:
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(*data)
                sector_time.time = data[0]
                sector_time.sector = data[1]
                sector_time.invalid = data[2]
                instance.add_sector_time(sector_time)

        self.assertEqual(instance.best_sector_1, expected_value)

    def test_property_best_sector_1_all_invalid(self):
        instance = Driver(sentinel.index, sentinel.name)
        test_data = [
            (30.000, 1, True),
            (45.000, 2, False),
            (40.000, 3, False),
            (33.000, 1, True),
            (43.000, 2, False),
            (39.000, 3, False)
        ]
        expected_value = None
        for data in test_data:
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(*data)
                sector_time.time = data[0]
                sector_time.sector = data[1]
                sector_time.invalid = data[2]
                instance.add_sector_time(sector_time)

        self.assertEqual(instance.best_sector_1, expected_value)

    def test_property_best_sector_2_default(self):
        instance = Driver(sentinel.index, sentinel.name)
        self.assertIsNone(instance.best_sector_2)

    def test_property_best_sector_2_valid(self):
        instance = Driver(sentinel.index, sentinel.name)
        test_data = [
            (30.000, 1, False),
            (45.000, 2, False),
            (40.000, 3, False),
            (33.000, 1, False),
            (43.000, 2, False),
            (39.000, 3, False)
        ]
        expected_value = 43.000
        for data in test_data:
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(*data)
                sector_time.time = data[0]
                sector_time.sector = data[1]
                sector_time.invalid = data[2]
                instance.add_sector_time(sector_time)

        self.assertEqual(instance.best_sector_2, expected_value)

    def test_property_best_sector_2_invalid(self):
        instance = Driver(sentinel.index, sentinel.name)
        test_data = [
            (30.000, 1, False),
            (45.000, 2, False),
            (40.000, 3, False),
            (33.000, 1, False),
            (43.000, 2, True),
            (39.000, 3, False)
        ]
        expected_value = 45.000
        for data in test_data:
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(*data)
                sector_time.time = data[0]
                sector_time.sector = data[1]
                sector_time.invalid = data[2]
                instance.add_sector_time(sector_time)

        self.assertEqual(instance.best_sector_2, expected_value)

    def test_property_best_sector_2_all_invalid(self):
        instance = Driver(sentinel.index, sentinel.name)
        test_data = [
            (30.000, 1, False),
            (45.000, 2, True),
            (40.000, 3, False),
            (33.000, 1, False),
            (43.000, 2, True),
            (39.000, 3, False)
        ]
        expected_value = None
        for data in test_data:
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(*data)
                sector_time.time = data[0]
                sector_time.sector = data[1]
                sector_time.invalid = data[2]
                instance.add_sector_time(sector_time)

        self.assertEqual(instance.best_sector_2, expected_value)

    def test_property_best_sector_3_default(self):
        instance = Driver(sentinel.index, sentinel.name)
        self.assertIsNone(instance.best_sector_3)

    def test_property_best_sector_3_valid(self):
        instance = Driver(sentinel.index, sentinel.name)
        test_data = [
            (30.000, 1, False),
            (45.000, 2, False),
            (40.000, 3, False),
            (33.000, 1, False),
            (43.000, 2, False),
            (39.000, 3, False)
        ]
        expected_value = 39.000
        for data in test_data:
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(*data)
                sector_time.time = data[0]
                sector_time.sector = data[1]
                sector_time.invalid = data[2]
                instance.add_sector_time(sector_time)

        self.assertEqual(instance.best_sector_3, expected_value)

    def test_property_best_sector_3_invalid(self):
        instance = Driver(sentinel.index, sentinel.name)
        test_data = [
            (30.000, 1, False),
            (45.000, 2, False),
            (40.000, 3, False),
            (33.000, 1, False),
            (43.000, 2, False),
            (39.000, 3, True)
        ]
        expected_value = 40.000
        for data in test_data:
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(*data)
                sector_time.time = data[0]
                sector_time.sector = data[1]
                sector_time.invalid = data[2]
                instance.add_sector_time(sector_time)

        self.assertEqual(instance.best_sector_3, expected_value)

    def test_property_best_sector_3_all_invalid(self):
        instance = Driver(sentinel.index, sentinel.name)
        test_data = [
            (30.000, 1, False),
            (45.000, 2, False),
            (40.000, 3, True),
            (33.000, 1, False),
            (43.000, 2, False),
            (39.000, 3, True)
        ]
        expected_value = None
        for data in test_data:
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(*data)
                sector_time.time = data[0]
                sector_time.sector = data[1]
                sector_time.invalid = data[2]
                instance.add_sector_time(sector_time)

        self.assertEqual(instance.best_sector_3, expected_value)

    def test_property_index(self):
        instance = Driver(sentinel.index, sentinel.name)
        expected_result = sentinel.index
        self.assertEqual(instance.index, expected_result)

    def test_property_index_setter(self):
        instance = Driver(sentinel.index, sentinel.name)
        instance.index = sentinel.new_index
        expected_result = sentinel.new_index
        self.assertEqual(instance.index, expected_result)

    def test_property_laps_complete(self):
        instance = Driver(sentinel.index, sentinel.name)
        times = [
            sentinel.time_1,
            sentinel.time_2,
            sentinel.time_3,
            sentinel.time_4,
            sentinel.time_5,
            sentinel.time_6
        ]
        for sector, time in enumerate(times, 1):
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(
                    time,
                    sector % 3,
                    False)
                sector_time.sector = sector % 3
                sector_time.time = time
                sector_time.invalid = False

                instance.add_sector_time(sector_time)

        expected_result = 2
        self.assertEqual(instance.laps_complete, expected_result)

    def test_property_laps_complete_empty(self):
        instance = Driver(sentinel.index, sentinel.name)
        expected_result = 0
        self.assertEqual(instance.laps_complete, expected_result)

    def test_property_lap_times(self):
        instance = Driver(sentinel.index, sentinel.name)
        times = [
            5,
            5,
            5,
            10,
            10,
            10
        ]
        for sector, time in enumerate(times, 1):
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(
                    time,
                    sector % 3,
                    False)
                sector_time.sector = sector % 3
                sector_time.time = time
                sector_time.invalid = False

                instance.add_sector_time(sector_time)

        expected_result = [15, 30]
        self.assertListEqual(instance.lap_times, expected_result)

    def test_property_lap_times_empty(self):
        instance = Driver(sentinel.index, sentinel.name)
        expected_result = list()
        self.assertListEqual(instance.lap_times, expected_result)

    def test_property_last_lap_invalid(self):
        instance = Driver(sentinel.index, sentinel.name)
        sectors = [
            (sentinel.time_l1s1, 1, False),
            (sentinel.time_l1s2, 2, False),
            (sentinel.time_l1s3, 3, False),
            (sentinel.time_l2s1, 1, False),
            (sentinel.time_l2s2, 2, False),
            (sentinel.time_l2s2, 2, True),
            (sentinel.time_l2s3, 3, True)
        ]

        for time, sector, invalid in sectors:
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(time, sector, invalid)
                sector_time.sector = sector
                sector_time.time = time
                sector_time.invalid = True if instance._invalidate_next_sector_count > 0 else invalid

                instance.add_sector_time(sector_time)

        self.assertTrue(instance.last_lap_invalid)

    def test_property_last_lap_invalid_previous_invalid(self):
        instance = Driver(sentinel.index, sentinel.name)
        sectors = [
            (sentinel.time_l1s1, 1, False),
            (sentinel.time_l1s2, 2, False),
            (sentinel.time_l1s2, 2, True),
            (sentinel.time_l1s3, 3, True),
            (sentinel.time_l2s1, 1, False),
            (sentinel.time_l2s2, 2, False),
            (sentinel.time_l2s3, 3, False)
        ]

        for time, sector, invalid in sectors:
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(time, sector, invalid)
                sector_time.sector = sector
                sector_time.time = time
                sector_time.invalid = invalid
                instance.add_sector_time(sector_time)

        self.assertFalse(instance.last_lap_invalid)

    def test_property_last_lap_invalid_no_laps(self):
        instance = Driver(sentinel.index, sentinel.name)
        self.assertIsNone(instance.last_lap_invalid)

    def test_property_last_lap_time(self):
        instance = Driver(sentinel.index, sentinel.name)
        times = [
            5,
            5,
            5,
            10,
            10,
            10
        ]
        for sector, time in enumerate(times, 1):
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(
                    time,
                    sector % 3,
                    False)
                sector_time.sector = sector % 3
                sector_time.time = time
                sector_time.invalid = False

                instance.add_sector_time(sector_time)

        expected_result = 30
        self.assertEqual(instance.last_lap_time, expected_result)

    def test_property_last_lap_time_no_laps(self):
        instance = Driver(sentinel.index, sentinel.name)
        self.assertIsNone(instance.last_lap_time)

    def test_property_name(self):
        instance = Driver(sentinel.index, sentinel.name)
        expected_result = sentinel.name
        self.assertEqual(instance.name, expected_result)

    def test_property_name_setter(self):
        instance = Driver(sentinel.index, sentinel.name)
        instance.name = sentinel.new_name
        expected_result = sentinel.new_name
        self.assertEqual(instance.name, expected_result)

    def test_property_race_time(self):
        instance = Driver(sentinel.index, sentinel.name)
        times = [
            5,
            5,
            5,
            10,
            10,
            10
        ]
        for sector, time in enumerate(times, 1):
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(
                    time,
                    sector % 3,
                    False)
                sector_time.sector = sector % 3
                sector_time.time = time
                sector_time.invalid = False

                instance.add_sector_time(sector_time)

        expected_result = 45
        self.assertEqual(instance.race_time, expected_result)

    def test_property_race_time_no_laps(self):
        instance = Driver(sentinel.index, sentinel.name)
        expected_result = 0.0
        self.assertEqual(instance.race_time, expected_result)

    def test_property_sector_times(self):
        instance = Driver(sentinel.index, sentinel.name)
        expected_result = list
        self.assertIsInstance(instance.sector_times, expected_result)

    def test_property_sector_times_values(self):
        instance = Driver(sentinel.index, sentinel.name)
        times = [
            sentinel.time_1,
            sentinel.time_2,
            sentinel.time_3,
            sentinel.time_4,
            sentinel.time_5,
            sentinel.time_6
        ]
        expected_value = list()
        for sector, time in enumerate(times, 1):
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(
                    time,
                    sector % 3,
                    False)
                sector_time.sector = sector % 3
                sector_time.time = time
                sector_time.invalid = False

                instance.add_sector_time(sector_time)
                expected_value.append(sector_time)

        self.assertListEqual(instance.sector_times, expected_value)

    @patch('replayenhancer.RaceData.SectorTime', autospec=True)
    def test_method_add_sector_time_no_previous(self, sector_time):
        instance = Driver(sentinel.index, sentinel.name)
        sector_time.time = -123.0
        sector_time.sector = 1
        sector_time.invalid = False

        instance.add_sector_time(sector_time)

        self.assertEqual(len(instance.sector_times), 0)

    @patch('replayenhancer.RaceData.SectorTime', autospec=True)
    def test_method_add_sector_time_first(self, sector_time):
        instance = Driver(sentinel.index, sentinel.name)
        sector_time.time = sentinel.time
        sector_time.sector = 1
        sector_time.invalid = False

        instance.add_sector_time(sector_time)

        self.assertEqual(len(instance.sector_times), 1)

    @patch('replayenhancer.RaceData.SectorTime', autospec=True)
    def test_method_add_sector_time_duplicate(self, sector_time):
        instance = Driver(sentinel.index, sentinel.name)
        sector_time.time = sentinel.time
        sector_time.sector = 1
        sector_time.invalid = False

        instance.add_sector_time(sector_time)
        instance.add_sector_time(sector_time)

        self.assertEqual(len(instance.sector_times), 1)

    def test_method_add_sector_time_invalidate_first_lap(self):
        instance = Driver(sentinel.index, sentinel.name)
        sectors = [
            (sentinel.time_s1, 1, False),
            (sentinel.time_s2, 2, False),
            (sentinel.time_s2, 2, True),
            (sentinel.time_s3, 3, True)
        ]

        for time, sector, invalid in sectors:
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(time, sector, invalid)
                sector_time.sector = sector
                sector_time.time = time
                sector_time.invalid = invalid

                instance.add_sector_time(sector_time)

        self.assertEqual(len(instance.sector_times), 3)
        self.assertTrue(all([sector_time.invalid for sector_time in instance.sector_times]))

    def test_method_add_sector_time_invalidate_later_lap(self):
        instance = Driver(sentinel.index, sentinel.name)
        sectors = [
            (sentinel.time_l1s1, 1, False),
            (sentinel.time_l1s2, 2, False),
            (sentinel.time_l1s3, 3, False),
            (sentinel.time_l2s1, 1, False),
            (sentinel.time_l2s2, 2, False),
            (sentinel.time_l2s2, 2, True),
            (sentinel.time_l2s3, 3, True)
        ]

        for time, sector, invalid in sectors:
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(time, sector, invalid)
                sector_time.sector = sector
                sector_time.time = time
                sector_time.invalid = invalid
                instance.add_sector_time(sector_time)

        # First lap should still be valid.
        result = [
            sector_time.invalid
            for sector_time in instance.sector_times[:3]]
        self.assertTrue(not any(result))

        # Second lap should be invalid.
        result = [
            sector_time.invalid
            for sector_time in instance.sector_times[3:]]
        self.assertTrue(all(result))

    def test_method_add_sector_time_invalidate_earlier_lap(self):
        instance = Driver(sentinel.index, sentinel.name)
        sectors = [
            (sentinel.time_l1s1, 1, False),
            (sentinel.time_l1s2, 2, False),
            (sentinel.time_l1s2, 2, True),
            (sentinel.time_l1s3, 3, True),
            (sentinel.time_l2s1, 1, False),
            (sentinel.time_l2s2, 2, False),
            (sentinel.time_l2s3, 3, False)
        ]

        for time, sector, invalid in sectors:
            with patch(
                    'replayenhancer.RaceData.SectorTime',
                    autospec=True) as SectorTime:
                sector_time = SectorTime(time, sector, invalid)
                sector_time.sector = sector
                sector_time.time = time
                sector_time.invalid = invalid
                instance.add_sector_time(sector_time)

        # First lap should be invalid.
        result = [
            sector_time.invalid
            for sector_time in instance.sector_times[:3]]
        self.assertTrue(any(result))

        # Second lap should be invalid.
        result = [
            sector_time.invalid
            for sector_time in instance.sector_times[3:]]
        self.assertTrue(not all(result))


class TestSectorTime(unittest.TestCase):
    """
    Unit tests for SectorTime object.
    """
    def setUp(self):
        self.instance = SectorTime(sentinel.time, sentinel.sector, 0)

    def tearDown(self):
        self.instance = None

    def test_init(self):
        expected_result = SectorTime
        self.assertIsInstance(self.instance, expected_result)

    def test_property_sector(self):
        expected_result = sentinel.sector
        self.assertEqual(self.instance.sector, expected_result)

    def test_property_time(self):
        expected_result = sentinel.time
        self.assertEqual(self.instance.time, expected_result)

    def test_property_invalid_false(self):
        self.assertFalse(self.instance.invalid)

    def test_property_invalid_true(self):
        instance = SectorTime(sentinel.time, sentinel.sector, 128)
        self.assertTrue(instance.invalid)

    def test_property_invalid_setter(self):
        self.instance.invalid = sentinel.new_invalid
        expected_result = sentinel.new_invalid
        self.assertEqual(self.instance.invalid, expected_result)

if __name__ == "__main__":
    unittest.main()



import os.path
import sys

import moviepy.editor as mpy
from numpy import diff, nonzero
from tqdm import tqdm

from replayenhancer.RaceData import RaceData

source_video = mpy.VideoFileClip(os.path.abspath(sys.argv[1]))
threshold = 1
gap_time = 1

blackframes = [
    t for (t, f)
    in source_video.iter_frames(with_times=True, progress_bar=1)
    if f.mean() < threshold]

print([blackframes[x] for x in nonzero(diff(blackframes)>gap_time)[0]])
print([blackframes[::-1][x] for x in nonzero(diff(blackframes[::-1])<-gap_time)[0]])

race_data = RaceData(os.path.abspath(sys.argv[2]))

with tqdm(desc="Processing telemetry") as progress:
    while True:
        try:
            _ = race_data.get_data()
            progress.update()
        except StopIteration:
            break

print(race_data.elapsed_time)



from replayenhancer.RaceData import TelemetryData
import sys

data = TelemetryData(sys.argv[1])

packet = next(data)
while packet.packet_type != 0:
    packet = next(data)

print("Track Length: {}".format(packet.track_length))

while packet.packet_type != 1:
    packet = next(data)

print("Track: {} / {}".format(packet.track_location, packet.track_variation))



"""Tests AdditionalParticipantPacket.py

"""
import unittest
from hashlib import md5
from struct import pack

from replayenhancer.AdditionalParticipantPacket \
    import AdditionalParticipantPacket


class TestAdditionalParticipantPacket(unittest.TestCase):
    """Unit tests against the AdditionalParticipantPacket object.

    """
    expected_build_version_number = 12345
    expected_packet_type = 2

    expected_offset = 16

    expected_name = [
        "Felipe Nasr",
        "Jolyon Palmer",
        "Pascal Wehrlein",
        "Stoffel Vandoorne",
        "Esteban Gutiérrez",
        "Marcus Ericsson",
        "Esteban Ocon",
        "Rio Haryanto",
        "",
        "",
        "",
        "",
        "",
        "",
        "",
        ""
    ]

    expected_packet_length = 1028

    @classmethod
    def binary_data(cls, **kwargs):
        test_data = list()
        packet_string = "HBB"
        packet_string += "64s"*16

        try:
            test_data.append(kwargs['build_version_number'])
        except KeyError:
            test_data.append(cls.expected_build_version_number)

        try:
            test_data.append(kwargs['packet_type'])
        except KeyError:
            test_data.append(cls.expected_packet_type)

        try:
            test_data.append(kwargs['offset'])
        except KeyError:
            test_data.append(cls.expected_offset)

        try:
            test_data.extend([name.encode('utf-8') for name in kwargs['name']])
        except KeyError:
            test_data.extend(
                [name.encode('utf-8') for name in cls.expected_name])

        return pack(packet_string, *test_data)

    def test_init(self):
        instance = AdditionalParticipantPacket(self.binary_data())
        expected_result = AdditionalParticipantPacket
        self.assertIsInstance(instance, expected_result)

    def test_init_wrong_packet_length(self):
        test_binary_data = pack("H", 42)

        from struct import error
        with self.assertRaises(error):
            AdditionalParticipantPacket(test_binary_data)

    def test_init_wrong_packet_type(self):
        with self.assertRaises(ValueError):
            AdditionalParticipantPacket(self.binary_data(packet_type=1))

    def test_property_build_version_number(self):
        instance = AdditionalParticipantPacket(self.binary_data())
        expected_result = self.expected_build_version_number
        self.assertEqual(instance.build_version_number, expected_result)

    def test_property_data_hash(self):
        instance = AdditionalParticipantPacket(self.binary_data())
        expected_result = md5(self.binary_data()).hexdigest()
        self.assertEqual(instance.data_hash, expected_result)

    def test_property_name(self):
        instance = AdditionalParticipantPacket(self.binary_data())
        expected_result = self.expected_name
        self.assertListEqual(instance.name, expected_result)

    def test_property_name_split_on_null(self):
        instance = AdditionalParticipantPacket(self.binary_data(
            name=[name+'\x00Garbage Data' for name in self.expected_name]))
        expected_result = self.expected_name
        self.assertListEqual(instance.name, expected_result)

    def test_property_offset(self):
        instance = AdditionalParticipantPacket(self.binary_data())
        expected_result = self.expected_offset
        self.assertEqual(instance.offset, expected_result)

    def test_property_packet_type(self):
        instance = AdditionalParticipantPacket(self.binary_data())
        expected_result = self.expected_packet_type
        self.assertEqual(instance.packet_type, expected_result)

    def test_method_repr(self):
        instance = AdditionalParticipantPacket(self.binary_data())
        expected_result = "AdditionalParticipantPacket"
        self.assertEqual(repr(instance), expected_result)

    def test_method_str(self):
        instance = AdditionalParticipantPacket(self.binary_data())
        expected_result = "AdditionalParticipantPacket"
        self.assertEqual(str(instance), expected_result)



"""Tests TelemetryDataPacket.py

"""
from unittest.mock import patch
import unittest
from struct import pack
from hashlib import md5
from collections import deque

from replayenhancer.TelemetryDataPacket \
    import ParticipantInfo, TelemetryDataPacket


class TestTelemetryDataPacket(unittest.TestCase):
    """Unit tests against the TelemetryDataPacket object.

    """
    from random import uniform, randrange
    expected_build_version_number = randrange(2000)
    expected_packet_type = 0

    expected_game_state = randrange(4)
    expected_session_state = randrange(7)

    expected_viewed_participant_index = randrange(57)
    expected_num_participants = randrange(57)

    expected_unfiltered_throttle = randrange(256)
    expected_unfiltered_brake = randrange(256)
    expected_unfiltered_steering = randrange(-128, 128)
    expected_unfiltered_clutch = randrange(256)
    expected_race_state = randrange(7)
    expected_lap_invalidated = randrange(2)
    expected_anti_lock = randrange(2)
    expected_boost = randrange(2)

    expected_laps_in_event = randrange(256)

    expected_best_lap_time = uniform(0, 360)
    expected_last_lap_time = uniform(0, 360)
    expected_current_time = uniform(0, 360)
    expected_split_time_ahead = uniform(0, 360)
    expected_split_time_behind = uniform(0, 360)
    expected_split_time = uniform(0, 360)
    expected_event_time_remaining = uniform(0, 360)
    expected_personal_fastest_lap_time = uniform(0, 360)
    expected_world_fastest_lap_time = uniform(0, 360)
    expected_current_s1_time = uniform(0, 360)
    expected_current_s2_time = uniform(0, 360)
    expected_current_s3_time = uniform(0, 360)
    expected_fastest_s1_time = uniform(0, 360)
    expected_fastest_s2_time = uniform(0, 360)
    expected_fastest_s3_time = uniform(0, 360)
    expected_personal_fastest_s1_time = uniform(0, 360)
    expected_personal_fastest_s2_time = uniform(0, 360)
    expected_personal_fastest_s3_time = uniform(0, 360)
    expected_world_fastest_s1_time = uniform(0, 360)
    expected_world_fastest_s2_time = uniform(0, 360)
    expected_world_fastest_s3_time = uniform(0, 360)

    expected_joypad = randrange(9)

    expected_highest_flag = randrange(65)

    expected_pit_mode_schedule = randrange(65)

    expected_oil_temp = randrange(-32767, 32767)
    expected_oil_pressure = randrange(65535)
    expected_water_temp = randrange(-32767, 32767)
    expected_water_pressure = randrange(65535)
    expected_fuel_pressure = randrange(65535)
    expected_car_flags = randrange(256)
    expected_fuel_capacity = randrange(256)
    expected_brake = randrange(256)
    expected_throttle = randrange(256)
    expected_clutch = randrange(256)
    expected_steering = randrange(-128, 128)
    expected_fuel_level = uniform(0, 256)
    expected_speed = uniform(0, 256)
    expected_rpm = randrange(65535)
    expected_max_rpm = randrange(65535)
    expected_gear_num_gears = randrange(256)
    expected_boost_amount = randrange(256)
    expected_enforced_pit_stop_lap = randrange(-128, 128)
    expected_crash_state = randrange(256)

    expected_odometer = uniform(0, 256)
    expected_orientation = [uniform(0, 360), uniform(0, 360), uniform(0, 360)]
    expected_local_velocity = [
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255)]
    expected_world_velocity = [
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255)]
    expected_angular_velocity = [
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255)]
    expected_local_acceleration = [
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255)]
    expected_world_acceleration = [
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255)]
    expected_extents_centre = [
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255)]

    expected_tyre_flags = [
        randrange(256),
        randrange(256),
        randrange(256),
        randrange(256)]
    expected_terrain = [
        randrange(256),
        randrange(256),
        randrange(256),
        randrange(256)]
    expected_tyre_y = [
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255)]
    expected_tyre_rps = [
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255)]
    expected_tyre_slip_speed = [
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255)]
    expected_tyre_temp = [
        randrange(256),
        randrange(256),
        randrange(256),
        randrange(256)]
    expected_tyre_grip = [
        randrange(256),
        randrange(256),
        randrange(256),
        randrange(256)]
    expected_tyre_height_above_ground = [
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255)]
    expected_tyre_lateral_stiffness = [
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255)]
    expected_tyre_wear = [
        randrange(256),
        randrange(256),
        randrange(256),
        randrange(256)]
    expected_brake_damage = [
        randrange(256),
        randrange(256),
        randrange(256),
        randrange(256)]
    expected_suspension_damage = [
        randrange(256),
        randrange(256),
        randrange(256),
        randrange(256)]
    expected_brake_temp = [
        randrange(-32767, 32767),
        randrange(-32767, 32767),
        randrange(-32767, 32767),
        randrange(-32767, 32767)]
    expected_tyre_tread_temp = [
        randrange(65535),
        randrange(65535),
        randrange(65535),
        randrange(65535)]
    expected_tyre_layer_temp = [
        randrange(65535),
        randrange(65535),
        randrange(65535),
        randrange(65535)]
    expected_tyre_carcass_temp = [
        randrange(65535),
        randrange(65535),
        randrange(65535),
        randrange(65535)]
    expected_tyre_rim_temp = [
        randrange(65535),
        randrange(65535),
        randrange(65535),
        randrange(65535)]
    expected_tyre_internal_air_temp = [
        randrange(65535),
        randrange(65535),
        randrange(65535),
        randrange(65535)]
    expected_wheel_local_position_y = [
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255)]
    expected_ride_height = [
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255)]
    expected_suspension_travel = [
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255)]
    expected_suspension_velocity = [
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255),
        uniform(0, 255)]
    expected_air_pressure = [
        randrange(65535),
        randrange(65535),
        randrange(65535),
        randrange(65535)]

    expected_engine_speed = uniform(0, 256)
    expected_engine_torque = uniform(0, 256)

    expected_aero_damage = randrange(256)
    expected_engine_damage = randrange(256)

    expected_ambient_temperature = randrange(-128, 128)
    expected_track_temperature = randrange(-128, 128)
    expected_rain_density = randrange(256)
    expected_wind_speed = randrange(-128, 128)
    expected_wind_direction_x = randrange(-128, 128)
    expected_wind_direction_y = randrange(-128, 128)

    expected_track_length = uniform(0, 256)
    expected_wings = [randrange(128), randrange(128)]
    expected_d_pad = randrange(128)

    def assertListAlmostEqual(self, list1, list2, delta):
        for item1, item2 in zip(list1, list2):
            self.assertAlmostEqual(item1, item2, delta=delta)

    @classmethod
    def binary_data(cls, **kwargs):
        test_data = list()
        packet_string = "HB"
        packet_string += "B"
        packet_string += "bb"
        packet_string += "BBbBB"
        packet_string += "B"
        packet_string += "21f"
        packet_string += "H"
        packet_string += "B"
        packet_string += "B"
        packet_string += "hHhHHBBBBBbffHHBBbB"
        packet_string += "22f"
        packet_string += "8B12f8B8f12B4h20H16f4H"
        packet_string += "2f"
        packet_string += "2B"
        packet_string += "bbBbbb"

        packet_string += "hhhHBBBBf"*56

        packet_string += "fBBB"

        try:
            test_data.append(kwargs['build_version_number'])
        except KeyError:
            test_data.append(cls.expected_build_version_number)

        try:
            test_data.append(kwargs['packet_type'])
        except KeyError:
            test_data.append(cls.expected_packet_type)

        try:
            game_state = kwargs['game_state']
        except KeyError:
            game_state = cls.expected_game_state

        try:
            session_state = kwargs['session_state']
        except KeyError:
            session_state = cls.expected_session_state

        test_data.append((session_state << 4) + game_state)

        try:
            test_data.append(kwargs['viewed_participant_index'])
        except KeyError:
            test_data.append(cls.expected_viewed_participant_index)

        try:
            test_data.append(kwargs['num_participants'])
        except KeyError:
            test_data.append(cls.expected_num_participants)

        try:
            test_data.append(kwargs['unfiltered_throttle'])
        except KeyError:
            test_data.append(cls.expected_unfiltered_throttle)

        try:
            test_data.append(kwargs['unfiltered_brake'])
        except KeyError:
            test_data.append(cls.expected_unfiltered_brake)

        try:
            test_data.append(kwargs['unfiltered_steering'])
        except KeyError:
            test_data.append(cls.expected_unfiltered_steering)

        try:
            test_data.append(kwargs['unfiltered_clutch'])
        except KeyError:
            test_data.append(cls.expected_unfiltered_clutch)

        try:
            race_state = kwargs['race_state']
        except KeyError:
            race_state = cls.expected_race_state

        try:
            lap_invalidated = kwargs['lap_invalidated']
        except KeyError:
            lap_invalidated = cls.expected_lap_invalidated

        try:
            anti_lock = kwargs['anti_lock']
        except KeyError:
            anti_lock = cls.expected_anti_lock

        try:
            boost = kwargs['boost']
        except KeyError:
            boost = cls.expected_boost

        test_data.append(
            (boost << 5)
            + (anti_lock << 4)
            + (lap_invalidated << 3)
            + race_state)

        try:
            test_data.append(kwargs['laps_in_event'])
        except KeyError:
            test_data.append(cls.expected_laps_in_event)

        try:
            test_data.append(kwargs['best_lap_time'])
        except KeyError:
            test_data.append(cls.expected_best_lap_time)

        try:
            test_data.append(kwargs['last_lap_time'])
        except KeyError:
            test_data.append(cls.expected_last_lap_time)

        try:
            test_data.append(kwargs['current_time'])
        except KeyError:
            test_data.append(cls.expected_current_time)

        try:
            test_data.append(kwargs['split_time_ahead'])
        except KeyError:
            test_data.append(cls.expected_split_time_ahead)

        try:
            test_data.append(kwargs['split_time_behind'])
        except KeyError:
            test_data.append(cls.expected_split_time_behind)

        try:
            test_data.append(kwargs['split_time'])
        except KeyError:
            test_data.append(cls.expected_split_time)

        try:
            test_data.append(kwargs['event_time_remaining'])
        except KeyError:
            test_data.append(cls.expected_event_time_remaining)

        try:
            test_data.append(kwargs['personal_fastest_lap_time'])
        except KeyError:
            test_data.append(cls.expected_personal_fastest_lap_time)

        try:
            test_data.append(kwargs['world_fastest_lap_time'])
        except KeyError:
            test_data.append(cls.expected_world_fastest_lap_time)

        try:
            test_data.append(kwargs['current_s1_time'])
        except KeyError:
            test_data.append(cls.expected_current_s1_time)

        try:
            test_data.append(kwargs['current_s2_time'])
        except KeyError:
            test_data.append(cls.expected_current_s2_time)

        try:
            test_data.append(kwargs['current_s3_time'])
        except KeyError:
            test_data.append(cls.expected_current_s3_time)

        try:
            test_data.append(kwargs['fastest_s1_time'])
        except KeyError:
            test_data.append(cls.expected_fastest_s1_time)

        try:
            test_data.append(kwargs['fastest_s2_time'])
        except KeyError:
            test_data.append(cls.expected_fastest_s2_time)

        try:
            test_data.append(kwargs['fastest_s3_time'])
        except KeyError:
            test_data.append(cls.expected_fastest_s3_time)

        try:
            test_data.append(kwargs['personal_fastest_s1_time'])
        except KeyError:
            test_data.append(cls.expected_personal_fastest_s1_time)

        try:
            test_data.append(kwargs['personal_fastest_s2_time'])
        except KeyError:
            test_data.append(cls.expected_personal_fastest_s2_time)

        try:
            test_data.append(kwargs['personal_fastest_s3_time'])
        except KeyError:
            test_data.append(cls.expected_personal_fastest_s3_time)

        try:
            test_data.append(kwargs['world_fastest_s1_time'])
        except KeyError:
            test_data.append(cls.expected_world_fastest_s1_time)

        try:
            test_data.append(kwargs['world_fastest_s2_time'])
        except KeyError:
            test_data.append(cls.expected_world_fastest_s2_time)

        try:
            test_data.append(kwargs['world_fastest_s3_time'])
        except KeyError:
            test_data.append(cls.expected_world_fastest_s3_time)

        try:
            test_data.append(kwargs['joypad'])
        except KeyError:
            test_data.append(cls.expected_joypad)

        try:
            test_data.append(kwargs['highest_flag'])
        except KeyError:
            test_data.append(cls.expected_highest_flag)

        try:
            test_data.append(kwargs['pit_mode_schedule'])
        except KeyError:
            test_data.append(cls.expected_pit_mode_schedule)

        try:
            test_data.append(kwargs['oil_temp'])
        except KeyError:
            test_data.append(cls.expected_oil_temp)

        try:
            test_data.append(kwargs['oil_pressure'])
        except KeyError:
            test_data.append(cls.expected_oil_pressure)

        try:
            test_data.append(kwargs['water_temp'])
        except KeyError:
            test_data.append(cls.expected_water_temp)

        try:
            test_data.append(kwargs['water_pressure'])
        except KeyError:
            test_data.append(cls.expected_water_pressure)

        try:
            test_data.append(kwargs['fuel_pressure'])
        except KeyError:
            test_data.append(cls.expected_fuel_pressure)

        try:
            test_data.append(kwargs['car_flags'])
        except KeyError:
            test_data.append(cls.expected_car_flags)

        try:
            test_data.append(kwargs['fuel_capacity'])
        except KeyError:
            test_data.append(cls.expected_fuel_capacity)

        try:
            test_data.append(kwargs['brake'])
        except KeyError:
            test_data.append(cls.expected_brake)

        try:
            test_data.append(kwargs['throttle'])
        except KeyError:
            test_data.append(cls.expected_throttle)

        try:
            test_data.append(kwargs['clutch'])
        except KeyError:
            test_data.append(cls.expected_clutch)

        try:
            test_data.append(kwargs['steering'])
        except KeyError:
            test_data.append(cls.expected_steering)

        try:
            test_data.append(kwargs['fuel_level'])
        except KeyError:
            test_data.append(cls.expected_fuel_level)

        try:
            test_data.append(kwargs['speed'])
        except KeyError:
            test_data.append(cls.expected_speed)

        try:
            test_data.append(kwargs['rpm'])
        except KeyError:
            test_data.append(cls.expected_rpm)

        try:
            test_data.append(kwargs['max_rpm'])
        except KeyError:
            test_data.append(cls.expected_max_rpm)

        try:
            test_data.append(kwargs['gear_num_gears'])
        except KeyError:
            test_data.append(cls.expected_gear_num_gears)

        try:
            test_data.append(kwargs['boost_amount'])
        except KeyError:
            test_data.append(cls.expected_boost_amount)

        try:
            test_data.append(kwargs['enforced_pit_stop_lap'])
        except KeyError:
            test_data.append(cls.expected_enforced_pit_stop_lap)

        try:
            test_data.append(kwargs['crash_state'])
        except KeyError:
            test_data.append(cls.expected_crash_state)

        try:
            test_data.append(kwargs['odometer'])
        except KeyError:
            test_data.append(cls.expected_odometer)

        try:
            test_data.extend(kwargs['orientation'])
        except KeyError:
            test_data.extend(cls.expected_orientation)

        try:
            test_data.extend(kwargs['local_velocity'])
        except KeyError:
            test_data.extend(cls.expected_local_velocity)

        try:
            test_data.extend(kwargs['world_velocity'])
        except KeyError:
            test_data.extend(cls.expected_world_velocity)

        try:
            test_data.extend(kwargs['angular_velocity'])
        except KeyError:
            test_data.extend(cls.expected_angular_velocity)

        try:
            test_data.extend(kwargs['local_acceleration'])
        except KeyError:
            test_data.extend(cls.expected_local_acceleration)

        try:
            test_data.extend(kwargs['world_acceleration'])
        except KeyError:
            test_data.extend(cls.expected_world_acceleration)

        try:
            test_data.extend(kwargs['extents_centre'])
        except KeyError:
            test_data.extend(cls.expected_extents_centre)

        try:
            test_data.extend(kwargs['tyre_flags'])
        except KeyError:
            test_data.extend(cls.expected_tyre_flags)

        try:
            test_data.extend(kwargs['terrain'])
        except KeyError:
            test_data.extend(cls.expected_terrain)

        try:
            test_data.extend(kwargs['tyre_y'])
        except KeyError:
            test_data.extend(cls.expected_tyre_y)

        try:
            test_data.extend(kwargs['tyre_rps'])
        except KeyError:
            test_data.extend(cls.expected_tyre_rps)

        try:
            test_data.extend(kwargs['tyre_slip_speed'])
        except KeyError:
            test_data.extend(cls.expected_tyre_slip_speed)

        try:
            test_data.extend(kwargs['tyre_temp'])
        except KeyError:
            test_data.extend(cls.expected_tyre_temp)

        try:
            test_data.extend(kwargs['tyre_grip'])
        except KeyError:
            test_data.extend(cls.expected_tyre_grip)

        try:
            test_data.extend(kwargs['tyre_height_above_ground'])
        except KeyError:
            test_data.extend(cls.expected_tyre_height_above_ground)

        try:
            test_data.extend(kwargs['tyre_lateral_stiffness'])
        except KeyError:
            test_data.extend(cls.expected_tyre_lateral_stiffness)

        try:
            test_data.extend(kwargs['tyre_wear'])
        except KeyError:
            test_data.extend(cls.expected_tyre_wear)

        try:
            test_data.extend(kwargs['brake_damage'])
        except KeyError:
            test_data.extend(cls.expected_brake_damage)

        try:
            test_data.extend(kwargs['suspension_damage'])
        except KeyError:
            test_data.extend(cls.expected_suspension_damage)

        try:
            test_data.extend(kwargs['brake_temp'])
        except KeyError:
            test_data.extend(cls.expected_brake_temp)

        try:
            test_data.extend(kwargs['tyre_tread_temp'])
        except KeyError:
            test_data.extend(cls.expected_tyre_tread_temp)

        try:
            test_data.extend(kwargs['tyre_layer_temp'])
        except KeyError:
            test_data.extend(cls.expected_tyre_layer_temp)

        try:
            test_data.extend(kwargs['tyre_carcass_temp'])
        except KeyError:
            test_data.extend(cls.expected_tyre_carcass_temp)

        try:
            test_data.extend(kwargs['tyre_rim_temp'])
        except KeyError:
            test_data.extend(cls.expected_tyre_rim_temp)

        try:
            test_data.extend(kwargs['tyre_internal_air_temp'])
        except KeyError:
            test_data.extend(cls.expected_tyre_internal_air_temp)

        try:
            test_data.extend(kwargs['wheel_local_position_y'])
        except KeyError:
            test_data.extend(cls.expected_wheel_local_position_y)

        try:
            test_data.extend(kwargs['ride_height'])
        except KeyError:
            test_data.extend(cls.expected_ride_height)

        try:
            test_data.extend(kwargs['suspension_travel'])
        except KeyError:
            test_data.extend(cls.expected_suspension_travel)

        try:
            test_data.extend(kwargs['suspension_velocity'])
        except KeyError:
            test_data.extend(cls.expected_suspension_velocity)

        try:
            test_data.extend(kwargs['air_pressure'])
        except KeyError:
            test_data.extend(cls.expected_air_pressure)

        try:
            test_data.append(kwargs['engine_speed'])
        except KeyError:
            test_data.append(cls.expected_engine_speed)

        try:
            test_data.append(kwargs['engine_torque'])
        except KeyError:
            test_data.append(cls.expected_engine_torque)

        try:
            test_data.append(kwargs['aero_damage'])
        except KeyError:
            test_data.append(cls.expected_aero_damage)

        try:
            test_data.append(kwargs['engine_damage'])
        except KeyError:
            test_data.append(cls.expected_engine_damage)

        try:
            test_data.append(kwargs['ambient_temperature'])
        except KeyError:
            test_data.append(cls.expected_ambient_temperature)

        try:
            test_data.append(kwargs['track_temperature'])
        except KeyError:
            test_data.append(cls.expected_track_temperature)

        try:
            test_data.append(kwargs['rain_density'])
        except KeyError:
            test_data.append(cls.expected_rain_density)

        try:
            test_data.append(kwargs['wind_speed'])
        except KeyError:
            test_data.append(cls.expected_wind_speed)

        try:
            test_data.append(kwargs['wind_direction_x'])
        except KeyError:
            test_data.append(cls.expected_wind_direction_x)

        try:
            test_data.append(kwargs['wind_direction_y'])
        except KeyError:
            test_data.append(cls.expected_wind_direction_y)

        # Add 896 bytes of pointless data to account for ParticipantInfo.
        for _ in range(56):
            test_data.append(1)
            test_data.append(1)
            test_data.append(1)
            test_data.append(2)
            test_data.append(3)
            test_data.append(3)
            test_data.append(3)
            test_data.append(3)
            test_data.append(4.0)

        try:
            test_data.append(kwargs['track_length'])
        except KeyError:
            test_data.append(cls.expected_track_length)

        try:
            test_data.extend(kwargs['wings'])
        except KeyError:
            test_data.extend(cls.expected_wings)

        try:
            test_data.append(kwargs['d_pad'])
        except KeyError:
            test_data.append(cls.expected_d_pad)

        return pack(packet_string, *test_data)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_init(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = TelemetryDataPacket
        self.assertIsInstance(instance, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_init_wrong_packet_length(self, _):
        test_binary_data = pack("H", 42)

        from struct import error
        with self.assertRaises(error):
            TelemetryDataPacket(test_binary_data)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_init_wrong_packet_type(self, _):
        with self.assertRaises(ValueError):
            TelemetryDataPacket(self.binary_data(packet_type=1))

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_build_version_number(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_build_version_number
        self.assertEqual(instance.build_version_number, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_data_hash(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = md5(self.binary_data()).hexdigest()
        self.assertEqual(instance.data_hash, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_game_state(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_game_state
        self.assertEqual(instance.game_state, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_session_state(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_session_state
        self.assertEqual(instance.session_state, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_viewed_participant_index(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_viewed_participant_index
        self.assertEqual(instance.viewed_participant_index, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_num_participants(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_num_participants
        self.assertEqual(instance.num_participants, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_unfiltered_throttle(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_unfiltered_throttle
        self.assertEqual(instance.unfiltered_throttle, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_unfiltered_brake(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_unfiltered_brake
        self.assertEqual(instance.unfiltered_brake, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_unfiltered_steering(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_unfiltered_steering
        self.assertEqual(instance.unfiltered_steering, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_unfiltered_clutch(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_unfiltered_clutch
        self.assertEqual(instance.unfiltered_clutch, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_race_state(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_race_state
        self.assertEqual(instance.race_state, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_laps_in_event(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_laps_in_event
        self.assertEqual(instance.laps_in_event, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_best_lap_time(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_best_lap_time
        self.assertAlmostEqual(instance.best_lap_time, expected_result, 4)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_last_lap_time(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_last_lap_time
        self.assertAlmostEqual(
            instance.last_lap_time,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_current_time(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_current_time
        self.assertAlmostEqual(instance.current_time, expected_result, delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_split_time_ahead(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_split_time_ahead
        self.assertAlmostEqual(
            instance.split_time_ahead,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_split_time_behind(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_split_time_behind
        self.assertAlmostEqual(
            instance.split_time_behind,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_split_time(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_split_time
        self.assertAlmostEqual(instance.split_time, expected_result, delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_event_time_remaining(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_event_time_remaining
        self.assertAlmostEqual(
            instance.event_time_remaining,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_personal_fastest_lap_time(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_personal_fastest_lap_time
        self.assertAlmostEqual(
            instance.personal_fastest_lap_time,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_world_fastest_lap_time(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_world_fastest_lap_time
        self.assertAlmostEqual(
            instance.world_fastest_lap_time,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_current_s1_time(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_current_s1_time
        self.assertAlmostEqual(
            instance.current_s1_time,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_current_s2_time(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_current_s2_time
        self.assertAlmostEqual(
            instance.current_s2_time,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_current_s3_time(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_current_s3_time
        self.assertAlmostEqual(
            instance.current_s3_time,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_fastest_s1_time(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_fastest_s1_time
        self.assertAlmostEqual(
            instance.fastest_s1_time,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_fastest_s2_time(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_fastest_s2_time
        self.assertAlmostEqual(
            instance.fastest_s2_time,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_fastest_s3_time(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_fastest_s3_time
        self.assertAlmostEqual(
            instance.fastest_s3_time,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_personal_fastest_s1_time(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_personal_fastest_s1_time
        self.assertAlmostEqual(
            instance.personal_fastest_s1_time,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_personal_fastest_s2_time(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_personal_fastest_s2_time
        self.assertAlmostEqual(
            instance.personal_fastest_s2_time,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_personal_fastest_s3_time(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_personal_fastest_s3_time
        self.assertAlmostEqual(
            instance.personal_fastest_s3_time,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_world_fastest_s1_time(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_world_fastest_s1_time
        self.assertAlmostEqual(
            instance.world_fastest_s1_time,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_world_fastest_s2_time(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_world_fastest_s2_time
        self.assertAlmostEqual(
            instance.world_fastest_s2_time,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_world_fastest_s3_time(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_world_fastest_s3_time
        self.assertAlmostEqual(
            instance.world_fastest_s3_time,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_joypad(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_joypad
        self.assertEqual(instance.joypad, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_highest_flag(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_highest_flag
        self.assertEqual(instance.highest_flag, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_pit_mode_schedule(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_pit_mode_schedule
        self.assertEqual(instance.pit_mode_schedule, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_oil_temp(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_oil_temp
        self.assertEqual(instance.oil_temp, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_oil_pressure(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_oil_pressure
        self.assertEqual(instance.oil_pressure, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_water_temp(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_water_temp
        self.assertEqual(instance.water_temp, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_water_pressure(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_water_pressure
        self.assertEqual(instance.water_pressure, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_fuel_pressure(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_fuel_pressure
        self.assertEqual(instance.fuel_pressure, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_car_flags(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_car_flags
        self.assertEqual(instance.car_flags, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_fuel_capacity(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_fuel_capacity
        self.assertEqual(instance.fuel_capacity, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_brake(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_brake
        self.assertEqual(instance.brake, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_throttle(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_throttle
        self.assertEqual(instance.throttle, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_clutch(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_clutch
        self.assertEqual(instance.clutch, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_steering(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_steering
        self.assertEqual(instance.steering, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_fuel_level(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_fuel_level
        self.assertAlmostEqual(instance.fuel_level, expected_result, delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_speed(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_speed
        self.assertAlmostEqual(instance.speed, expected_result, delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_rpm(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_rpm
        self.assertEqual(instance.rpm, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_max_rpm(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_max_rpm
        self.assertEqual(instance.max_rpm, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_gear_num_gears(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_gear_num_gears
        self.assertEqual(instance.gear_num_gears, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_boost_amount(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_boost_amount
        self.assertEqual(instance.boost_amount, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_enforced_pit_stop_lap(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_enforced_pit_stop_lap
        self.assertEqual(instance.enforced_pit_stop_lap, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_crash_state(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_crash_state
        self.assertEqual(instance.crash_state, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_odometer(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_odometer
        self.assertAlmostEqual(instance.odometer, expected_result, delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_orientation(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_orientation
        self.assertListAlmostEqual(
            instance.orientation,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_local_velocity(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_local_velocity
        self.assertListAlmostEqual(
            instance.local_velocity,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_world_velocity(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_world_velocity
        self.assertListAlmostEqual(
            instance.world_velocity,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_angular_velocity(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_angular_velocity
        self.assertListAlmostEqual(
            instance.angular_velocity,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_local_acceleration(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_local_acceleration
        self.assertListAlmostEqual(
            instance.local_acceleration,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_world_acceleration(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_world_acceleration
        self.assertListAlmostEqual(
            instance.world_acceleration,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_extents_centre(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_extents_centre
        self.assertListAlmostEqual(
            instance.extents_centre,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_tyre_flags(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_tyre_flags
        self.assertListEqual(instance.tyre_flags, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_terrain(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_terrain
        self.assertListEqual(instance.terrain, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_tyre_y(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_tyre_y
        self.assertListAlmostEqual(instance.tyre_y, expected_result, delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_tyre_rps(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_tyre_rps
        self.assertListAlmostEqual(instance.tyre_rps, expected_result, delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_tyre_slip_speed(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_tyre_slip_speed
        self.assertListAlmostEqual(
            instance.tyre_slip_speed,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_tyre_temp(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_tyre_temp
        self.assertListEqual(instance.tyre_temp, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_tyre_grip(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_tyre_grip
        self.assertListEqual(instance.tyre_grip, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_tyre_height_above_ground(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_tyre_height_above_ground
        self.assertListAlmostEqual(
            instance.tyre_height_above_ground,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_tyre_lateral_stiffness(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_tyre_lateral_stiffness
        self.assertListAlmostEqual(
            instance.tyre_lateral_stiffness,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_tyre_wear(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_tyre_wear
        self.assertListEqual(instance.tyre_wear, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_brake_damage(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_brake_damage
        self.assertListEqual(instance.brake_damage, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_suspension_damage(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_suspension_damage
        self.assertListEqual(instance.suspension_damage, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_brake_temp(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_brake_temp
        self.assertListEqual(instance.brake_temp, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_tyre_tread_temp(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_tyre_tread_temp
        self.assertListEqual(instance.tyre_tread_temp, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_tyre_layer_temp(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_tyre_layer_temp
        self.assertListEqual(instance.tyre_layer_temp, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_tyre_carcass_temp(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_tyre_carcass_temp
        self.assertListEqual(instance.tyre_carcass_temp, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_tyre_rim_temp(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_tyre_rim_temp
        self.assertListEqual(instance.tyre_rim_temp, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_tyre_internal_air_temp(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_tyre_internal_air_temp
        self.assertListEqual(instance.tyre_internal_air_temp, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_wheel_local_position_y(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_wheel_local_position_y
        self.assertListAlmostEqual(
            instance.wheel_local_position_y,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_ride_height(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_ride_height
        self.assertListAlmostEqual(
            instance.ride_height,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_suspension_travel(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_suspension_travel
        self.assertListAlmostEqual(
            instance.suspension_travel,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_suspension_velocity(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_suspension_velocity
        self.assertListAlmostEqual(
            instance.suspension_velocity,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_air_pressure(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_air_pressure
        self.assertListEqual(instance.air_pressure, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_engine_speed(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_engine_speed
        self.assertAlmostEqual(instance.engine_speed, expected_result, delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_engine_torque(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_engine_torque
        self.assertAlmostEqual(
            instance.engine_torque,
            expected_result,
            delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_aero_damage(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_aero_damage
        self.assertEqual(instance.aero_damage, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_engine_damage(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_engine_damage
        self.assertEqual(instance.engine_damage, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_ambient_temperature(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_ambient_temperature
        self.assertEqual(instance.ambient_temperature, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_track_temperature(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_track_temperature
        self.assertEqual(instance.track_temperature, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_rain_density(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_rain_density
        self.assertEqual(instance.rain_density, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_wind_speed(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_wind_speed
        self.assertEqual(instance.wind_speed, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_wind_direction_x(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_wind_direction_x
        self.assertEqual(instance.wind_direction_x, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_wind_direction_y(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_wind_direction_y
        self.assertEqual(instance.wind_direction_y, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_participant_info(self, _):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = list
        self.assertIsInstance(instance.participant_info, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_track_length(self, mock_participant_info):
        def remove_participant_info_data(*args):
            for _ in range(9):
                args[0].popleft()
        mock_participant_info.side_effect = remove_participant_info_data

        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_track_length
        self.assertAlmostEqual(instance.track_length, expected_result, delta=0.001)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_wings(self, mock_participant_info):
        def remove_participant_info_data(*args):
            for _ in range(9):
                args[0].popleft()
        mock_participant_info.side_effect = remove_participant_info_data

        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_wings
        self.assertListEqual(instance.wings, expected_result)

    @patch('replayenhancer.TelemetryDataPacket.ParticipantInfo', autospec=True)
    def test_property_d_pad(self, mock_participant_info):
        def remove_participant_info_data(*args):
            for _ in range(9):
                args[0].popleft()
        mock_participant_info.side_effect = remove_participant_info_data

        instance = TelemetryDataPacket(self.binary_data())
        expected_result = self.expected_d_pad
        self.assertEqual(instance.d_pad, expected_result)

    def test_method_repr(self):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = "TelemetryDataPacket"
        self.assertEqual(repr(instance), expected_result)

    def test_method_str(self):
        instance = TelemetryDataPacket(self.binary_data())
        expected_result = "TelemetryDataPacket"
        self.assertEqual(str(instance), expected_result)


class TestParticipantInfo(unittest.TestCase):
    """Unit tests against the ParticipantInfo object.

    """
    from random import choice, random, randrange
    expected_world_position = [
        randrange(100)+choice([0, 0.25, 0.50, 0.75]),
        randrange(100),
        randrange(100) + choice([0, 0.25, 0.50, 0.75])]
    expected_current_lap_distance = randrange(100)
    expected_is_active = 1
    expected_race_position = randrange(56)
    expected_invalid_lap = 0
    expected_laps_completed = randrange(50)
    expected_current_lap = expected_laps_completed + 1
    expected_sector = randrange(1, 4)
    expected_last_sector_time = random()

    @classmethod
    def binary_data(cls, **kwargs):
        test_data = list()

        try:
            world_position = kwargs['world_position']
        except KeyError:
            world_position = cls.expected_world_position

        test_data.extend([int(num) for num in world_position])

        try:
            test_data.append(kwargs['current_lap_distance'])
        except KeyError:
            test_data.append(cls.expected_current_lap_distance)

        try:
            is_active = kwargs['is_active']
        except KeyError:
            is_active = cls.expected_is_active

        try:
            race_position = kwargs['race_position']
        except KeyError:
            race_position = cls.expected_race_position

        test_data.append((is_active << 7) + race_position)

        try:
            invalid_lap = kwargs['invalid_lap']
        except KeyError:
            invalid_lap = cls.expected_invalid_lap

        try:
            laps_completed = kwargs['laps_completed']
        except KeyError:
            laps_completed = cls.expected_laps_completed

        test_data.append((invalid_lap << 7) + laps_completed)

        try:
            test_data.append(kwargs['current_lap'])
        except KeyError:
            test_data.append(cls.expected_current_lap)

        try:
            sector = kwargs['sector']
        except KeyError:
            sector = cls.expected_sector

        x_acc = int(((world_position[0] - int(world_position[0])) * 100) / 25)
        z_acc = int(((world_position[2] - int(world_position[2])) * 100) / 25)
        test_data.append((z_acc << 5) + (x_acc << 3) + sector)

        try:
            test_data.append(kwargs['last_sector_time'])
        except KeyError:
            test_data.append(cls.expected_last_sector_time)

        return deque(tuple(test_data))

    def test_init(self):
        instance = ParticipantInfo(self.binary_data())
        expected_result = ParticipantInfo
        self.assertIsInstance(instance, expected_result)

    def test_property_current_lap(self):
        instance = ParticipantInfo(self.binary_data())
        expected_result = self.expected_current_lap
        self.assertEqual(instance.current_lap, expected_result)

    def test_property_current_lap_distance(self):
        instance = ParticipantInfo(self.binary_data())
        expected_result = self.expected_current_lap_distance
        self.assertEqual(instance.current_lap_distance, expected_result)

    def test_property_invalid_lap_false(self):
        instance = ParticipantInfo(self.binary_data(invalid_lap=0))
        self.assertFalse(instance.invalid_lap)

    def test_property_invalid_lap_true(self):
        instance = ParticipantInfo(self.binary_data(invalid_lap=1))
        self.assertTrue(instance.invalid_lap)

    def test_property_invalid_lap_false_negative_at_race_start(self):
        instance = ParticipantInfo(self.binary_data(
            invalid_lap=1,
            sector=3,
            last_sector_time=-123))
        self.assertFalse(instance.invalid_lap)

    def test_property_is_active_false(self):
        instance = ParticipantInfo(self.binary_data(is_active=0))
        self.assertFalse(instance.is_active)

    def test_property_is_active_true(self):
        instance = ParticipantInfo(self.binary_data(is_active=1))
        self.assertTrue(instance.is_active)

    def test_property_laps_completed(self):
        instance = ParticipantInfo(self.binary_data())
        expected_result = self.expected_laps_completed
        self.assertEqual(instance.laps_completed, expected_result)

    def test_property_last_sector_time(self):
        instance = ParticipantInfo(self.binary_data())
        expected_result = self.expected_last_sector_time
        self.assertEqual(instance.last_sector_time, expected_result)

    def test_property_race_position(self):
        instance = ParticipantInfo(self.binary_data())
        expected_result = self.expected_race_position
        self.assertEqual(instance.race_position, expected_result)

    def test_property_sector(self):
        instance = ParticipantInfo(self.binary_data())
        expected_result = self.expected_sector
        self.assertEqual(instance.sector, expected_result)

    def test_property_world_position(self):
        instance = ParticipantInfo(self.binary_data())
        expected_result = self.expected_world_position
        self.assertListEqual(instance.world_position, expected_result)



"""
Provides a class for the Telemetry Data packets output by
Project CARS
"""

from hashlib import md5

from replayenhancer.Packet import Packet


class ParticipantInfo:
    """
    Creates an object containing the participant info from the
    telemetry data.
    """
    def __init__(self, unpacked_data):
        self._world_position = list()
        for _ in range(3):
            self._world_position.append(int(unpacked_data.popleft()))

        self.current_lap_distance = int(unpacked_data.popleft())
        self._race_position = int(unpacked_data.popleft())
        self._laps_completed = int(unpacked_data.popleft())
        self.current_lap = int(unpacked_data.popleft())
        self._sector = int(unpacked_data.popleft())
        self.last_sector_time = float(unpacked_data.popleft())

    @property
    def world_position(self):
        """Returns world position (high accuracy for x and z)."""
        world_position = [float(x) for x in self._world_position]
        world_position[0] += float(
            ((self._sector & int('00011000', 2)) >> 3) / 4)
        world_position[2] += float(
            ((self._sector & int('01100000', 2)) >> 5) / 4)

        return world_position

    @property
    def is_active(self):
        """Determines if the Participant is active."""
        return self._race_position & int('10000000', 2)

    @property
    def race_position(self):
        """Determines the Participant's race position."""
        return self._race_position & int('01111111', 2)

    @property
    def invalid_lap(self):
        """
        Determines if the Participant's lap is valid.

        Project CARS flags the start of the race (Sector 3, before you
        reach the start-finish line to begin the 'first lap proper')
        as invalid, so we need to deal with that. Dumb.

        That will have sector equal to 3 (since you start the race
        in Sector 3) and a previous sector time of -123 (since you
        don't have a previous sector time).
        """
        invalid_lap = self._laps_completed & int('10000000', 2)

        if invalid_lap and \
                self.sector == 3 and \
                self.last_sector_time == -123:
            return 0
        else:
            return invalid_lap

    @property
    def laps_completed(self):
        """Determines the laps completed by Participant."""
        return self._laps_completed & int('01111111', 2)

    @property
    def sector(self):
        """Determines the Participant's current sector."""
        return self._sector & int('00000111', 2)


class TelemetryDataPacket(Packet):
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-statements
    """
    Creates an object from a telemetry data packet.

    The telemetry data packet has a length of 1367 and is packet type
    0.
    """
    def __init__(self, packet_data):
        self.data_hash = md5(packet_data).hexdigest()
        unpacked_data = self._unpack_data(packet_data)

        self.build_version_number = int(unpacked_data.popleft())

        self._test_packet_type(unpacked_data.popleft())

        self._game_session_state = int(unpacked_data.popleft())

        self.viewed_participant_index = int(unpacked_data.popleft())
        self.num_participants = int(unpacked_data.popleft())

        self.unfiltered_throttle = int(unpacked_data.popleft())
        self.unfiltered_brake = int(unpacked_data.popleft())
        self.unfiltered_steering = int(unpacked_data.popleft())
        self.unfiltered_clutch = int(unpacked_data.popleft())
        self._race_state_flags = int(unpacked_data.popleft())

        self.laps_in_event = int(unpacked_data.popleft())

        self.best_lap_time = float(unpacked_data.popleft())
        self.last_lap_time = float(unpacked_data.popleft())
        self.current_time = float(unpacked_data.popleft())
        self.split_time_ahead = float(unpacked_data.popleft())
        self.split_time_behind = float(unpacked_data.popleft())
        self.split_time = float(unpacked_data.popleft())
        self.event_time_remaining = float(unpacked_data.popleft())
        self.personal_fastest_lap_time = float(
            unpacked_data.popleft())
        self.world_fastest_lap_time = float(unpacked_data.popleft())
        self.current_s1_time = float(unpacked_data.popleft())
        self.current_s2_time = float(unpacked_data.popleft())
        self.current_s3_time = float(unpacked_data.popleft())
        self.fastest_s1_time = float(unpacked_data.popleft())
        self.fastest_s2_time = float(unpacked_data.popleft())
        self.fastest_s3_time = float(unpacked_data.popleft())
        self.personal_fastest_s1_time = float(
            unpacked_data.popleft())
        self.personal_fastest_s2_time = float(
            unpacked_data.popleft())
        self.personal_fastest_s3_time = float(
            unpacked_data.popleft())
        self.world_fastest_s1_time = float(unpacked_data.popleft())
        self.world_fastest_s2_time = float(unpacked_data.popleft())
        self.world_fastest_s3_time = float(unpacked_data.popleft())

        self.joypad = int(unpacked_data.popleft())

        self.highest_flag = int(unpacked_data.popleft())

        self.pit_mode_schedule = int(unpacked_data.popleft())

        self.oil_temp = int(unpacked_data.popleft())
        self.oil_pressure = int(unpacked_data.popleft())
        self.water_temp = int(unpacked_data.popleft())
        self.water_pressure = int(unpacked_data.popleft())
        self.fuel_pressure = int(unpacked_data.popleft())
        self.car_flags = int(unpacked_data.popleft())
        self.fuel_capacity = int(unpacked_data.popleft())
        self.brake = int(unpacked_data.popleft())
        self.throttle = int(unpacked_data.popleft())
        self.clutch = int(unpacked_data.popleft())
        self.steering = int(unpacked_data.popleft())
        self.fuel_level = float(unpacked_data.popleft())
        self.speed = float(unpacked_data.popleft())
        self.rpm = int(unpacked_data.popleft())
        self.max_rpm = int(unpacked_data.popleft())
        self.gear_num_gears = int(unpacked_data.popleft())
        self.boost_amount = int(unpacked_data.popleft())
        self.enforced_pit_stop_lap = int(unpacked_data.popleft())
        self.crash_state = int(unpacked_data.popleft())

        self.odometer = float(unpacked_data.popleft())
        self.orientation = list()
        self.local_velocity = list()
        self.world_velocity = list()
        self.angular_velocity = list()
        self.local_acceleration = list()
        self.world_acceleration = list()
        self.extents_centre = list()
        for _ in range(3):
            self.orientation.append(float(
                unpacked_data.popleft()))
        for _ in range(3):
            self.local_velocity.append(float(
                unpacked_data.popleft()))
        for _ in range(3):
            self.world_velocity.append(float(
                unpacked_data.popleft()))
        for _ in range(3):
            self.angular_velocity.append(float(
                unpacked_data.popleft()))
        for _ in range(3):
            self.local_acceleration.append(float(
                unpacked_data.popleft()))
        for _ in range(3):
            self.world_acceleration.append(float(
                unpacked_data.popleft()))
        for _ in range(3):
            self.extents_centre.append(float(
                unpacked_data.popleft()))

        self.tyre_flags = list()
        self.terrain = list()
        self.tyre_y = list()
        self.tyre_rps = list()
        self.tyre_slip_speed = list()
        self.tyre_temp = list()
        self.tyre_grip = list()
        self.tyre_height_above_ground = list()
        self.tyre_lateral_stiffness = list()
        self.tyre_wear = list()
        self.brake_damage = list()
        self.suspension_damage = list()
        self.brake_temp = list()
        self.tyre_tread_temp = list()
        self.tyre_layer_temp = list()
        self.tyre_carcass_temp = list()
        self.tyre_rim_temp = list()
        self.tyre_internal_air_temp = list()
        self.wheel_local_position_y = list()
        self.ride_height = list()
        self.suspension_travel = list()
        self.suspension_velocity = list()
        self.air_pressure = list()

        for _ in range(4):
            self.tyre_flags.append(int(unpacked_data.popleft()))
        for _ in range(4):
            self.terrain.append(int(unpacked_data.popleft()))
        for _ in range(4):
            self.tyre_y.append(float(unpacked_data.popleft()))
        for _ in range(4):
            self.tyre_rps.append(float(unpacked_data.popleft()))
        for _ in range(4):
            self.tyre_slip_speed.append(float(
                unpacked_data.popleft()))
        for _ in range(4):
            self.tyre_temp.append(int(unpacked_data.popleft()))
        for _ in range(4):
            self.tyre_grip.append(int(unpacked_data.popleft()))
        for _ in range(4):
            self.tyre_height_above_ground.append(
                float(unpacked_data.popleft()))
        for _ in range(4):
            self.tyre_lateral_stiffness.append(float(
                unpacked_data.popleft()))
        for _ in range(4):
            self.tyre_wear.append(int(unpacked_data.popleft()))
        for _ in range(4):
            self.brake_damage.append(int(unpacked_data.popleft()))
        for _ in range(4):
            self.suspension_damage.append(int(
                unpacked_data.popleft()))
        for _ in range(4):
            self.brake_temp.append(int(unpacked_data.popleft()))
        for _ in range(4):
            self.tyre_tread_temp.append(int(
                unpacked_data.popleft()))
        for _ in range(4):
            self.tyre_layer_temp.append(int(
                unpacked_data.popleft()))
        for _ in range(4):
            self.tyre_carcass_temp.append(int(
                unpacked_data.popleft()))
        for _ in range(4):
            self.tyre_rim_temp.append(int(unpacked_data.popleft()))
        for _ in range(4):
            self.tyre_internal_air_temp.append(int(
                unpacked_data.popleft()))
        for _ in range(4):
            self.wheel_local_position_y.append(float(
                unpacked_data.popleft()))
        for _ in range(4):
            self.ride_height.append(float(unpacked_data.popleft()))
        for _ in range(4):
            self.suspension_travel.append(float(
                unpacked_data.popleft()))
        for _ in range(4):
            self.suspension_velocity.append(float(
                unpacked_data.popleft()))
        for _ in range(4):
            self.air_pressure.append(int(unpacked_data.popleft()))

        self.engine_speed = float(unpacked_data.popleft())
        self.engine_torque = float(unpacked_data.popleft())

        self.aero_damage = int(unpacked_data.popleft())
        self.engine_damage = int(unpacked_data.popleft())

        self.ambient_temperature = int(unpacked_data.popleft())
        self.track_temperature = int(unpacked_data.popleft())
        self.rain_density = int(unpacked_data.popleft())
        self.wind_speed = int(unpacked_data.popleft())
        self.wind_direction_x = int(unpacked_data.popleft())
        self.wind_direction_y = int(unpacked_data.popleft())

        self.participant_info = list()
        for _ in range(56):
            self.participant_info.append(ParticipantInfo(
                unpacked_data))

        self.track_length = float(unpacked_data.popleft())
        self.wings = list()
        for _ in range(2):
            self.wings.append(int(unpacked_data.popleft()))
        self.d_pad = int(unpacked_data.popleft())

    @property
    def packet_type(self):
        return 0

    @property
    def _packet_string(self):
        packet_string = "HB"
        packet_string += "B"
        packet_string += "bb"
        packet_string += "BBbBB"
        packet_string += "B"
        packet_string += "21f"
        packet_string += "H"
        packet_string += "B"
        packet_string += "B"
        packet_string += "hHhHHBBBBBbffHHBBbB"
        packet_string += "22f"
        packet_string += "8B12f8B8f12B4h20H16f4H"
        packet_string += "2f"
        packet_string += "2B"
        packet_string += "bbBbbb"

        packet_string += "hhhHBBBBf"*56

        packet_string += "fBBB"

        return packet_string

    def __str__(self):
        return "TelemetryDataPacket"

    @property
    def game_state(self):
        """Extracts and returns the game state."""
        return self._game_session_state & int('00001111', 2)

    @property
    def session_state(self):
        """Extracts and returns the session state."""
        return (self._game_session_state & int('11110000', 2)) >> 4

    @property
    def race_state(self):
        """Extracts and returns the race state."""
        return self._race_state_flags & int('00000111', 2)



"""
Provides classes for the reading and processing of captured Project
CARS telemetry data.
"""
import json
import os.path
from glob import glob
from hashlib import md5
from itertools import tee

from natsort import natsorted
from tqdm import tqdm

from replayenhancer.AdditionalParticipantPacket \
    import AdditionalParticipantPacket \
    as AdditionalParticipantPacket
from replayenhancer.REParticipantPacket \
    import REParticipantPacket \
    as ParticipantPacket
from replayenhancer.RETelemetryDataPacket \
    import RETelemetryDataPacket \
    as TelemetryDataPacket
from replayenhancer.StartingGridEntry import StartingGridEntry
from replayenhancer.Track import Track


class RaceData:
    """
    Holds data regarding the race.
    """
    def __init__(self, telemetry_directory, *,
                 descriptor_filename='descriptor.json'):
        self._starting_grid = list()
        self.drivers = dict()
        self._dropped_drivers = dict()

        self._stopped_drivers = set()

        self.elapsed_time = 0.0
        self._last_packet = None
        self._next_packet = None

        self.track = None

        self._descriptor_filename = descriptor_filename
        self._telemetry_directory = telemetry_directory
        self.telemetry_data = TelemetryData(
            telemetry_directory,
            descriptor_filename=descriptor_filename)

        self.get_data()

    @property
    def all_driver_classification(self):
        classification = self.classification
        for driver in self._dropped_drivers.values():
            classification.append(ClassificationEntry(None, driver, False))

        position = 0
        for entry in sorted(
                classification,
                key=lambda x: (-x.driver.laps_complete, x.driver.race_time)):
            position += 1
            entry.position = position

        return classification

    @property
    def best_lap(self):
        try:
            return min([
                driver.best_lap for driver
                in self.drivers.values()
                if driver.best_lap is not None])
        except ValueError:
            return None

    @property
    def best_sector_1(self):
        return self._best_sector(1)

    @property
    def best_sector_2(self):
        return self._best_sector(2)

    @property
    def best_sector_3(self):
        return self._best_sector(3)

    @property
    def classification(self):
        """
        Returns classification data at the current time.
        """
        classification = list()
        drivers_by_index = sorted(
            [driver for driver in self.drivers.values()],
            key=lambda x: x.index)
        for index in range(self._next_packet.num_participants):
            classification.append(ClassificationEntry(
                self._next_packet.participant_info[index].race_position,
                drivers_by_index[index],
                self._next_packet.viewed_participant_index == index))

        return classification

    @property
    def current_lap(self):
        leader_lap = max(
            [
                participant.current_lap for participant
                in self._next_packet.participant_info])
        return min(leader_lap, self.total_laps)

    @property
    def race_state(self):
        """
        Returns the current race state.
        0: RACESTATE_INVALID
        1: RACESTATE_NOT_STARTED
        2: RACESTATE_RACING,
        3: RACESTATE_FINISHED,
        4: RACESTATE_DISQUALIFIED,
        5: RACESTATE_RETIRED,
        6: RACESTATE_DNF,
        7: RACESTATE_MAX
        """
        return self._next_packet.race_state

    @property
    def starting_grid(self):
        """
        Returns the starting grid for the race.
        """
        if len(self._starting_grid):
            return self._starting_grid
        else:
            grid_data = TelemetryData(
                self._telemetry_directory,
                descriptor_filename=self._descriptor_filename)
            packet = None

            while packet is None or packet.packet_type != 0:
                packet = next(grid_data)

            drivers = sorted(
                self._get_drivers(
                    grid_data,
                    packet.num_participants).values(),
                key=lambda x: x.index)

            self._starting_grid = [
                StartingGridEntry(
                    participant_info.race_position,
                    index,
                    drivers[index].name)
                if index < len(drivers) else None
                for index, participant_info
                in enumerate(packet.participant_info)]

            self._starting_grid = [
                entry for entry in self._starting_grid
                if entry is not None]

            return self._starting_grid

    @property
    def total_laps(self):
        return self._next_packet.laps_in_event

    def driver_world_position(self, index):
        return self._next_packet.participant_info[index].world_position

    def get_data(self, at_time=None):
        """
        Retrieves the next telemetry packet.
        """
        while True:
            self._last_packet = self._next_packet
            self._next_packet = None

            try:
                while self._next_packet is None \
                        or self._next_packet.packet_type != 0:
                    self._next_packet = next(self.telemetry_data)
            except StopIteration:
                self._next_packet = self._last_packet
                raise

            if (self._next_packet is not None
                    and self._last_packet is None) \
                    or self._next_packet.num_participants \
                    != self._last_packet.num_participants:
                data, restore = tee(self.telemetry_data, 2)

                current_drivers = self._get_drivers(
                    data,
                    self._next_packet.num_participants)
                del data
                self.telemetry_data = restore

                # Add any new drivers.
                for key in current_drivers.keys() - self.drivers.keys():
                    self.drivers[key] = current_drivers[key]

                # Delete any dropped drivers.
                for key in self.drivers.keys() - current_drivers.keys():
                    self._dropped_drivers[key] = self.drivers[key]
                    del self.drivers[key]

                # Reset indices for drivers that remain.
                for key in current_drivers.keys():
                    self.drivers[key].index = current_drivers[key].index

            self.track = Track(self._next_packet.track_length)
            self._add_sector_times(self._next_packet)
            self._calc_elapsed_time()

            if at_time is None or self.elapsed_time >= at_time:
                return self._next_packet

    def _add_sector_times(self, packet):
        for index, participant_info in enumerate(
                packet.participant_info[:packet.num_participants]):
            driver_name = None
            for driver in self.drivers.values():
                if driver.index == index:
                    driver_name = driver.name

            if participant_info.sector == 1:
                sector = 3
            elif participant_info.sector == 2:
                if driver_name in self._stopped_drivers:
                    self._stopped_drivers.remove(driver_name)
                sector = 1
            elif participant_info.sector == 3:
                sector = 2
            else:
                """
                TODO: Investigate instance of a driver existing but having an
                invalid sector number (0). I suspect it's due to a network
                timeout.
                """
                return

            if self.track.at_pit_entry(participant_info.world_position) \
                    and driver_name not in self._stopped_drivers \
                    and self.race_state == 2:
                self._stopped_drivers.add(driver_name)

            if self.track.at_pit_exit(participant_info.world_position) \
                    and driver_name in self._stopped_drivers:
                self.drivers[driver_name].stops += 1
                self._stopped_drivers.remove(driver_name)

            sector_time = SectorTime(
                participant_info.last_sector_time,
                sector,
                participant_info.invalid_lap)

            self.drivers[driver_name].add_sector_time(sector_time)

    def _best_sector(self, sector):
        try:
            if sector == 1:
                return min(
                    [driver.best_sector_1 for driver in self.drivers.values()])
            elif sector == 2:
                return min(
                    [driver.best_sector_2 for driver in self.drivers.values()])
            elif sector == 3:
                return min(
                    [driver.best_sector_3 for driver in self.drivers.values()])
            else:
                raise ValueError
        except ValueError:
            return None

    def _calc_elapsed_time(self):
        if self._next_packet.current_time == -1.0:
            self.elapsed_time = 0.0
            self._last_packet = None
        else:
            driver = next(
                driver for driver in self.drivers.values()
                if driver.index == self._next_packet.viewed_participant_index)
            self.elapsed_time = sum(driver.lap_times) + self._next_packet.current_time

    @staticmethod
    def _get_drivers(telemetry_data, count):
        drivers = list()
        packet = next(telemetry_data)

        while len(drivers) < count:
            if packet.packet_type == 0 \
                    and packet.num_participants != count:
                raise ValueError(
                    "Participants not populated before break.")
            elif packet.packet_type == 1:
                for index, name in enumerate(packet.name):
                    drivers.append(Driver(index, name))
            elif packet.packet_type == 2:
                for index, name in enumerate(
                        packet.name,
                        packet.offset):
                    drivers.append(Driver(index, name))

            packet = next(telemetry_data)

        return {driver.name: driver for driver in sorted(
            drivers,
            key=lambda x: x.index)[:count]}


class ClassificationEntry:
    """
    Represents an entry on the classification table.
    """
    def __init__(self, position, driver, viewed_driver):
        self.position = position
        self.driver = driver
        self.viewed_driver = viewed_driver

    @property
    def best_lap(self):
        return self.driver.best_lap

    @property
    def best_sector_1(self):
        return self.driver.best_sector_1

    @property
    def best_sector_2(self):
        return self.driver.best_sector_2

    @property
    def best_sector_3(self):
        return self.driver.best_sector_3

    @property
    def driver_name(self):
        return self.driver.name

    @property
    def laps_complete(self):
        return self.driver.laps_complete

    @property
    def calc_points_data(self):
        return self.driver_name, self.position, self.best_lap

    @property
    def race_time(self):
        return self.driver.race_time

    @property
    def stops(self):
        return self.driver.stops


class Driver:
    """
    Represents a driver in the race.
    """
    _invalidate_next_sector_count = 0

    def __init__(self, index, name):
        self.index = index
        self.name = name

        self.sector_times = list()
        self.stops = 0

    @property
    def best_lap(self):
        valid_laps = list()
        for invalid, time in zip(
                self._lap_invalid(), self._lap_times()):
            if not invalid:
                valid_laps.append(time)

        try:
            return min(valid_laps)
        except ValueError:
            return None

    @property
    def best_sector_1(self):
        return self._best_sector(1)

    @property
    def best_sector_2(self):
        return self._best_sector(2)

    @property
    def best_sector_3(self):
        return self._best_sector(3)

    @property
    def laps_complete(self):
        return len(self.sector_times) // 3

    @property
    def lap_times(self):
        return self._lap_times()

    @property
    def last_lap_invalid(self):
        try:
            return self._lap_invalid()[-1]
        except IndexError:
            return None

    @property
    def last_lap_time(self):
        try:
            return self._lap_times()[-1]
        except IndexError:
            return None

    @property
    def race_time(self):
        return sum([sector_time.time for sector_time in self.sector_times])

    def add_sector_time(self, sector_time):
        if sector_time.time == -123.0:
            pass
        elif len(self.sector_times) == 0:
            self.sector_times.append(sector_time)
        elif self.sector_times[-1].time != sector_time.time \
                or self.sector_times[-1].invalid != sector_time.invalid \
                or self.sector_times[-1].sector != sector_time.sector:
            if self.sector_times[-1].invalid != sector_time.invalid \
                    and self.sector_times[-1].time == sector_time.time \
                    and self.sector_times[-1].sector == sector_time.sector:
                self.sector_times[-1] = sector_time
            else:
                if self._invalidate_next_sector_count > 0:
                    self.sector_times.append(SectorTime(
                        sector_time.time,
                        sector_time.sector,
                        True))
                    self._invalidate_next_sector_count -= 1
                else:
                    self.sector_times.append(SectorTime(
                        sector_time.time,
                        sector_time.sector,
                        False))

            if sector_time.invalid:
                self._invalidate_lap(sector_time)

    def _best_sector(self, sector):
        try:
            return min([
                sector_time.time
                for sector_time in self.sector_times
                if not sector_time.invalid
                and sector_time.sector == sector])
        except ValueError:
            return None

    def _invalidate_lap(self, sector_time):
        if sector_time.sector == 3:
            self._invalidate_next_sector_count = 3
        elif sector_time.sector == 1:
            self._invalidate_next_sector_count = 2
            for sector in self.sector_times[-1:]:
                sector.invalid = True
        elif sector_time.sector == 2:
            self._invalidate_next_sector_count = 1
            for sector in self.sector_times[-2:]:
                sector.invalid = True
        else:
            raise ValueError("Invalid Sector Number")

    def _lap_times(self):
        """
        Check to see if the first sector in the list is sector 1.
        Trim if not.
        """
        sector_times = self.sector_times
        try:
            while sector_times[0].sector != 1:
                sector_times = sector_times[1:]
        except IndexError:
            pass

        times = [sector.time for sector in sector_times]
        lap_times = list()
        for lap in zip(*[iter(times)]*3):
            try:
                lap_times.append(sum(lap))
            except TypeError:
                lap_times.append(None)
        return lap_times

    def _lap_invalid(self):
        """
        Check to see if the first sector in the list is sector 1.
        Trim if not.
        """
        sector_times = self.sector_times
        try:
            while sector_times[0].sector != 1:
                sector_times = sector_times[1:]
        except IndexError:
            pass

        invalids = [sector.invalid for sector in sector_times]
        lap_validity = list()
        for lap in zip(*[iter(invalids)]*3):
            try:
                lap_validity.append(any(lap))
            except TypeError:
                lap_validity.append(None)
        return lap_validity


class SectorTime:
    """
    Represents a sector time.
    """
    def __init__(self, time, sector, invalid):
        self.time = time
        self.sector = sector
        self.invalid = False if invalid == 0 else True


class TelemetryData:
    """
    Reads a directory of telemetry data and returns it as requested.
    """
    def __init__(self, telemetry_directory, *,
                 reverse=False,
                 descriptor_filename='descriptor.json'):
        if not os.path.isdir(telemetry_directory):
            raise NotADirectoryError

        self.packet_count = len(
            glob(telemetry_directory + os.sep + 'pdata*'))
        descriptor = None
        try:
            with open(os.path.join(
                os.path.realpath(telemetry_directory),
                os.path.relpath(descriptor_filename))) \
                    as descriptor_file:
                descriptor = json.load(descriptor_file)
        except (FileNotFoundError, ValueError):
            descriptor = self._build_descriptor(
                telemetry_directory,
                descriptor_filename)
        finally:
            self._telemetry_data = self._get_telemetry_data(
                telemetry_directory,
                descriptor,
                reverse=reverse)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._telemetry_data)

    def _build_descriptor(self, telemetry_directory, descriptor_filename):
        descriptor = {
            'race_end': None,
            'race_finish': None,
            'race_start': None}

        telemetry_data = self._get_telemetry_data(telemetry_directory)
        progress = tqdm(
            desc='Detecting Race End',
            total=self.packet_count,
            unit='packets')

        try:
            while True:
                while True:
                    packet = next(telemetry_data)
                    progress.update()
                    if packet.packet_type == 0 and packet.race_state == 3:
                        break

                # Exhaust packets until the race end.
                # TODO: Support for other ways to finish a race?
                while True:
                    try:
                        old_packet = packet
                        packet = next(telemetry_data)
                        progress.update()
                        if packet.packet_type == 0 and packet.race_state != 3:
                            break
                    except StopIteration:
                        old_packet = packet
                        break

        except StopIteration:
            progress.close()
            descriptor['race_end'] = old_packet.data_hash

        telemetry_data = self._get_telemetry_data(
            telemetry_directory,
            reverse=True)
        progress = tqdm(
            desc='Analyzing Telemetry Data',
            total=self.packet_count,
            unit='packets')

        # Exhaust packets until the race end.
        while True:
            packet = next(telemetry_data)
            progress.update()
            if packet.data_hash == descriptor['race_end']:
                break

        # Exhaust packets after the race finish.
        while True:
            packet = next(telemetry_data)
            progress.update()
            if packet.packet_type == 0 and packet.race_state == 2:
                break

        descriptor['race_finish'] = packet.data_hash

        # Exhaust packets after the green flag.
        while True:
            packet = next(telemetry_data)
            progress.update()
            if packet.packet_type == 0 and (
                    packet.race_state == 0 or
                    packet.race_state == 1):
                break

        old_packet = packet
        # Exhaust packets after the race start, except one.
        try:
            while True:
                if packet.packet_type == 0:
                    old_packet = packet
                packet = next(telemetry_data)
                progress.update()
                if packet.packet_type == 0 and (
                        packet.session_state != 5 or
                        packet.game_state != 2):
                    break
        # None found (no restarts?) so use the first packet.
        # It is already on `old_packet`.
        except StopIteration:
            pass

        progress.close()
        descriptor['race_start'] = old_packet.data_hash

        with open(os.path.join(
            os.path.realpath(telemetry_directory),
            os.path.relpath(descriptor_filename)), 'w') \
                as descriptor_file:
            json.dump(descriptor, descriptor_file)

        return descriptor

    @staticmethod
    def _get_telemetry_data(telemetry_directory, descriptor=None, *,
                            reverse=False):
        find_start = False if descriptor is None else True
        find_populate = False

        for packet in natsorted(
                glob(telemetry_directory+os.sep+'pdata*'),
                reverse=reverse):
            with open(packet, 'rb') as packet_file:
                packet_data = packet_file.read()

            if descriptor is not None \
                    and md5(packet_data).hexdigest() == descriptor['race_end']:
                raise StopIteration

            if find_start and \
                    md5(packet_data).hexdigest() == \
                    descriptor['race_start']:
                find_start = False
                find_populate = True
            elif find_start and \
                    md5(packet_data).hexdigest() != \
                    descriptor['race_start']:
                continue
            elif find_populate and \
                    len(packet_data) == 1367:
                packet = TelemetryDataPacket(packet_data)
                """
                TODO: Make sure this is actually correct. I think it's due
                to network lag during race loading.
                """
                positions = [
                    participant.race_position for participant
                    in packet.participant_info][:packet.num_participants]
                sectors = [
                    participant.sector for participant
                    in packet.participant_info][:packet.num_participants]

                if not all(positions) or not all(sectors):
                    continue
                else:
                    find_populate = False
                    yield packet
            elif len(packet_data) == 1367:
                yield TelemetryDataPacket(packet_data)
            elif len(packet_data) == 1347:
                yield ParticipantPacket(packet_data)
            elif len(packet_data) == 1028:
                yield AdditionalParticipantPacket(packet_data)



"""
Provides a class for the Telemetry Data packets output by
Project CARS.

Customized for use by the Project CARS Replay Enhancer.
"""

from hashlib import md5

from replayenhancer.TelemetryDataPacket import \
    ParticipantInfo, TelemetryDataPacket


class REParticipantInfo(ParticipantInfo):
    # pylint: disable=super-init-not-called
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-statements
    """
    Creates an object containing the participant info from the
    telemetry data.

    Customized for use by the Project CARS Replay Enhancer.
    We do not call the parent constructor.
    """
    def __init__(self, unpacked_data):
        self._world_position = list()
        for _ in range(3):
            self._world_position.append(int(unpacked_data.popleft()))
        self._race_position = int(unpacked_data.popleft())
        self._laps_completed = int(unpacked_data.popleft())
        self.current_lap = int(unpacked_data.popleft())
        self._sector = int(unpacked_data.popleft())
        self.last_sector_time = float(unpacked_data.popleft())


class RETelemetryDataPacket(TelemetryDataPacket):
    # pylint: disable=super-init-not-called
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-statements
    """
    Creates an object from a telemetry data packet.

    The telemetry data packet has a length of 1367 and is packet type
    0.

    Customized for use by the Project CARS Replay Enhancer.
    We do not call the parent constructor.
    """

    def __init__(self, packet_data):
        self.data_hash = md5(packet_data).hexdigest()
        unpacked_data = self._unpack_data(packet_data)

        self.build_version_number = int(unpacked_data.popleft())

        self._test_packet_type(unpacked_data.popleft())

        self._game_session_state = int(unpacked_data.popleft())

        self.viewed_participant_index = int(unpacked_data.popleft())
        self.num_participants = int(unpacked_data.popleft())

        self._race_state_flags = int(unpacked_data.popleft())

        self.laps_in_event = int(unpacked_data.popleft())

        self.current_time = float(unpacked_data.popleft())
        self.event_time_remaining = float(unpacked_data.popleft())

        self.participant_info = list()
        for _ in range(56):
            self.participant_info.append(REParticipantInfo(
                unpacked_data))
        self.participant_info[self.viewed_participant_index].\
            viewed = True

        self.track_length = float(unpacked_data.popleft())

    @property
    def _packet_string(self):
        """
        Original definition:
        _packet_string = "HB"
        _packet_string += "B"
        _packet_string += "bb"
        _packet_string += "BBbBB"
        _packet_string += "B"
        _packet_string += "21f"
        _packet_string += "H"
        _packet_string += "B"
        _packet_string += "B"
        _packet_string += "hHhHHBBBBBbffHHBBbB"
        _packet_string += "22f"
        _packet_string += "8B12f8B8f12B4h20H16f4H"
        _packet_string += "2f"
        _packet_string += "2B"
        _packet_string += "bbBbbb"

        _packet_string += "hhhHBBBBf"*56

        _packet_string += "fBBB"
        """
        packet_string = "HB"
        packet_string += "B"  # Game states
        packet_string += "bb"  # Participant info
        packet_string += "4xB"  # Unfiltered input
        packet_string += "B"  # Event Information
        packet_string += "8xf12xf56x"  # Timings
        packet_string += "2x"  # Joypad
        packet_string += "x"  # Flags
        packet_string += "x"  # Pit info
        packet_string += "10x6x8x4x4x"  # Car state
        packet_string += "88x"
        packet_string += "228x"  # Wheels / tyres
        packet_string += "8x"  # Extras
        packet_string += "2x"  # Car damage
        packet_string += "6x"  # Weather
        packet_string += "hhh2xBBBBf"*56  # Participant info
        packet_string += "fxxx"

        return packet_string

    def __str__(self):
        return "RETelemetryData"

    def set_name(self, name, index):
        """Sets the name for the specified index."""
        self.participant_info[index].name = name
        self.participant_info[index].viewed = \
            index == self.viewed_participant_index

    @property
    def event_duration(self):
        """Returns the event duration, in laps or time."""
        # TODO: Time
        return int(self.laps_in_event)

    @property
    def event_type(self):
        """Returns 'laps' or 'time'."""
        if isinstance(self.event_duration, int):
            return 'laps'
        else:
            return 'time'

    @property
    def drivers_by_position(self):
        """Returns drivers, sorted by race position."""
        return sorted(
            self.drivers_by_index,
            key=lambda x: x.race_position)

    @property
    def drivers_by_index(self):
        """Returns drivers, sorted by their index."""
        return [x for x in self.participant_info if x.is_active]\
            [:self.num_participants]

    @property
    def leader_lap(self):
        """Returns the leader's current lap."""
        return min(
            self.drivers_by_position[0].current_lap,
            self.event_duration)

    @property
    def last_place(self):
        """Returns the last place race position."""
        return max(
            [
                x.race_position
                for x in self.participant_info
                if x.is_active])



import json
import sys

config1 = json.load(open(sys.argv[1]))
config2 = json.load(open(sys.argv[2]))

expected_diffs = {
    'output_video',
    'source_telemetry',
    'source_video',
    'subheading_text',
    'sync_racestart',
    'video_skipend',
    'video_skipstart'}

expected_participant_diffs = {
    'points'}

for key in sorted(config1.keys() | config2.keys()):
    try:
        if key == 'participant_config' and config1[key] != config2[key]:
            participant_config1 = config1[key]
            participant_config2 = config2[key]

            common_drivers = sorted(participant_config1.keys() & participant_config2.keys())
            missing_drivers = sorted(participant_config1.keys() ^ participant_config2.keys())

            for driver in common_drivers:
                for driver_key in participant_config1[driver].keys():
                    if participant_config1[driver][driver_key] != participant_config2[driver][driver_key]:
                        if driver_key in expected_participant_diffs:
                            print("  ->EDIFF: {} {}".format(driver, driver_key))
                        else:
                            print("--->DIFF : {} {}".format(driver, driver_key))

            for driver in missing_drivers:
                print("--->NODRV: {}".format(driver))
                
        elif config1[key] == config2[key]:
            print("    EQUAL: {}".format(key))
        else:
            if key in expected_diffs:
                print("  ->EDIFF: {}".format(key))
            else:
                print("--->DIFF : {}".format(key))
    except KeyError:
        print("--->NOKEY: {}".format(key))
